#!/usr/bin/env python

from typing import Dict, List, Optional

import can

from pyxcp.cmdline import ArgumentParser
from pyxcp.transport.can import CanInterfaceBase


class CustomCANInterface(CanInterfaceBase):

    def init(self):
        """Initialize the CAN interface here."""

    def set_filters(self, filters):
        print(f"set_filters({filters})")
        self._filters = filters

    def recv(self, timeout: Optional[float] = None) -> Optional[can.message.Message]:
        """Receive CAN frames."""
        return can.message.Message()

    def send(self, msg: can.message.Message):
        """Send CAN frames."""
        print(f"send({msg})")

    @property
    def filters(self):
        """Return the current CAN filters."""
        return self._filters

    @property
    def state(self):
        """Return the current state of the CAN interface."""
        return can.BusState.ACTIVE


custom_interface = CustomCANInterface()

ap = ArgumentParser(description="User supplied CAN driver.")
with ap.run(transport_layer_interface=custom_interface) as x:
    x.connect()
    x.disconnect()



#!/usr/bin/env python
""" """

import functools
import operator
from abc import ABC, abstractmethod
from bisect import bisect_left
from enum import IntEnum
from typing import Any, Dict, List, Optional, Union

from can import (
    BusState,
    CanError,
    CanInitializationError,
    Message,
    detect_available_configs,
)
from can.bus import BusABC
from can.interface import _get_class_for_interface
from rich.console import Console

from pyxcp.config import CAN_INTERFACE_MAP
from pyxcp.transport.base import BaseTransport

from ..utils import seconds_to_nanoseconds


console = Console()

CAN_EXTENDED_ID = 0x80000000
MAX_11_BIT_IDENTIFIER = (1 << 11) - 1
MAX_29_BIT_IDENTIFIER = (1 << 29) - 1
MAX_DLC_CLASSIC = 8
CAN_FD_DLCS = (12, 16, 20, 24, 32, 48, 64)  # Discrete CAN-FD DLCs in case DLC > 8.


class FilterState(IntEnum):
    REJECT_ALL = 0
    ACCEPT_ALL = 1
    FILTERING = 2


class SoftwareFilter:
    """Additional CAN filters in software."""

    def __init__(self) -> None:
        self.filters = None
        self.reject_all()

    def set_filters(self, filters: List[Dict]) -> None:
        self.filters = filters
        self.filtering()

    def reject_all(self) -> None:
        self.filter_state = FilterState.REJECT_ALL

    def accept_all(self) -> None:
        self.filter_state = FilterState.ACCEPT_ALL

    def filtering(self) -> None:
        self.filter_state = FilterState.FILTERING

    @property
    def state(self) -> FilterState:
        return self.filter_state

    def accept(self, msg: Message) -> bool:
        """
        Based on: https://github.com/hardbyte/python-can/blob/bc248e8aaf96280a574c06e8e7d2778a67f091e3/can/bus.py#L430
        """
        if self.filter_state == FilterState.REJECT_ALL:
            return False
        elif self.filter_state == FilterState.ACCEPT_ALL or self.filters is None:
            return True
        for filter in self.filters:
            if "extended" in filter:
                if filter["extended"] != msg.is_extended_id:
                    continue
            can_id = filter["can_id"]
            can_mask = filter["can_mask"]
            if (can_id ^ msg.arbitration_id) & can_mask == 0:
                return True
        return False


class IdentifierOutOfRangeError(Exception):
    """Signals an identifier greater then :obj:`MAX_11_BIT_IDENTIFIER` or :obj:`MAX_29_BIT_IDENTIFIER`."""

    pass


def is_extended_identifier(identifier: int) -> bool:
    """Check for extendend CAN identifier.

    Parameters
    ----------
    identifier: int

    Returns
    -------
    bool
    """
    return (identifier & CAN_EXTENDED_ID) == CAN_EXTENDED_ID


def stripIdentifier(identifier: int) -> int:
    """Get raw CAN identifier (remove :obj:`CAN_EXTENDED_ID` bit if present).

    Parameters
    ----------
    identifier: int

    Returns
    -------
    int
    """
    return identifier & (~0xE0000000)


def samplePointToTsegs(tqs: int, samplePoint: float) -> tuple:
    """Calculate TSEG1 and TSEG2 from time-quantas and sample-point.

    Parameters
    ----------
    tqs: int
        Number of time-quantas
    samplePoint: float or int
        Sample-point as a percentage value.

    Returns
    -------
    tuple (TSEG1, TSEG2)
    """
    factor = samplePoint / 100.0
    tseg1 = int(tqs * factor)
    tseg2 = tqs - tseg1
    return (tseg1, tseg2)


def pad_frame(frame: bytes, pad_frame: bool, padding_value: int) -> bytes:
    """Pad frame to next discrete DLC value (CAN-FD) or on request (CAN-Classic).

    References:
    -----------
    ISO/DIS 15765 - 4; 8.2 Data length Code (DLC)
    AUTOSAR CP Release 4.3.0, Specification of CAN Transport Layer; 7.3.8 N-PDU padding
    AUTOSAR CP Release 4.3.0, Specification of CAN Driver; [SWS_CAN_00502], [ECUC_Can_00485]
    AUTOSAR CP Release 4.3.0, Requirements on CAN; [SRS_Can_01073], [SRS_Can_01086], [SRS_Can_01160]
    """
    frame_len = len(frame)
    if frame_len <= MAX_DLC_CLASSIC:
        actual_len = MAX_DLC_CLASSIC if pad_frame else frame_len
    else:
        actual_len = CAN_FD_DLCS[bisect_left(CAN_FD_DLCS, frame_len)]
    # append fill bytes up to MAX_DLC resp. next discrete FD DLC.
    if len(frame) < actual_len:
        frame += bytes([padding_value]) * (actual_len - len(frame))
    return frame


class Identifier:
    """Convenience class for XCP formatted CAN identifiers.

    Parameters:
    -----------
    raw_id: int
        Bit 32 set (i.e. 0x80000000) signals an extended (29-bit) identifier.

    Raises
    ------
    :class:`IdentifierOutOfRangeError`
    """

    def __init__(self, raw_id: int):
        self._raw_id = raw_id
        self._id = stripIdentifier(raw_id)
        self._is_extended = is_extended_identifier(raw_id)
        if self._is_extended:
            if self._id > MAX_29_BIT_IDENTIFIER:
                raise IdentifierOutOfRangeError(f"29-bit identifier {self._id!r} is out of range")
        else:
            if self._id > MAX_11_BIT_IDENTIFIER:
                raise IdentifierOutOfRangeError(f"11-bit identifier {self._id!r} is out of range")

    @property
    def id(self) -> int:
        """
        Returns
        -------
        int
            Identifier as seen on bus.
        """
        return self._id

    @property
    def raw_id(self) -> int:
        """
        Returns
        -------
        int
            Raw XCP formatted identifier.
        """
        return self._raw_id

    @property
    def is_extended(self) -> bool:
        """
        Returns
        -------
        bool
            - True - 29-bit identifier.
            - False - 11-bit identifier.
        """
        return self._is_extended

    @property
    def type_str(self) -> str:
        """

        Returns
        -------
        str
            - "S" - 11-bit identifier.
            - "E" - 29-bit identifier.
        """
        return "E" if self.is_extended else "S"

    @staticmethod
    def make_identifier(identifier: int, extended: bool) -> "Identifier":
        """Factory method.

        Parameters
        ----------
        identifier: int
            Identifier as seen on bus.

        extended: bool
            bool
                - True - 29-bit identifier.
                - False - 11-bit identifier.
        Returns
        -------
        :class:`Identifier`

        Raises
        ------
        :class:`IdentifierOutOfRangeError`
        """
        return Identifier(identifier if not extended else (identifier | CAN_EXTENDED_ID))

    def create_filter_from_id(self) -> Dict:
        """Create a single CAN filter entry.
        s. https://python-can.readthedocs.io/en/stable/bus.html#filtering
        """
        return {
            "can_id": self.id,
            "can_mask": MAX_29_BIT_IDENTIFIER if self.is_extended else MAX_11_BIT_IDENTIFIER,
            "extended": self.is_extended,
        }

    def __eq__(self, other) -> bool:
        return (self.id == other.id) and (self.is_extended == other.is_extended)

    def __str__(self) -> str:
        return f"Identifier(id = 0x{self.id:08x}, is_extended = {self.is_extended})"

    def __repr__(self) -> str:
        return f"Identifier(0x{self.raw_id:08x})"


class Frame:
    """"""

    def __init__(self, id_: Identifier, dlc: int, data: bytes, timestamp: int) -> None:
        self.id: Identifier = id_
        self.dlc: int = dlc
        self.data: bytes = data
        self.timestamp: int = timestamp

    def __repr__(self) -> str:
        return f"Frame(id = 0x{self.id:08x}, dlc = {self.dlc}, data = {self.data}, timestamp = {self.timestamp})"

    __str__ = __repr__


class PythonCanWrapper:
    """Wrapper around python-can - github.com/hardbyte/python-can"""

    def __init__(self, parent, interface_name: str, timeout: int, **parameters) -> None:
        self.parent = parent
        self.interface_name: str = interface_name
        self.timeout: int = timeout
        self.parameters = parameters
        if not self.parent.has_user_supplied_interface:
            self.can_interface_class = _get_class_for_interface(self.interface_name)
        else:
            self.can_interface_class = None
        self.can_interface: BusABC
        self.connected: bool = False
        self.software_filter = SoftwareFilter()
        self.saved_filters = []

    def connect(self) -> None:
        if self.connected:
            return
        can_filters = []
        can_filters.append(self.parent.can_id_slave.create_filter_from_id())  # Primary CAN filter.
        if self.parent.daq_identifier:
            # Add filters for DAQ identifiers.
            for daq_id in self.parent.daq_identifier:
                can_filters.append(daq_id.create_filter_from_id())
        if self.parent.has_user_supplied_interface:
            self.saved_filters = self.parent.transport_layer_interface.filters
            if self.saved_filters:
                merged_filters = can_filters[::]
                for fltr in self.saved_filters:
                    if fltr not in merged_filters:
                        merged_filters.append(fltr)
            self.can_interface = self.parent.transport_layer_interface
            self.can_interface.set_filters(merged_filters)
            self.software_filter.set_filters(can_filters)  # Filter unwanted traffic.
        else:
            self.can_interface = self.can_interface_class(interface=self.interface_name, can_filters=can_filters, **self.parameters)
            self.software_filter.accept_all()
        self.parent.logger.info(f"XCPonCAN - Using Interface: '{self.can_interface!s}'")
        self.parent.logger.info(f"XCPonCAN - Filters used: {self.can_interface.filters}")
        self.parent.logger.info(f"XCPonCAN - State: {self.can_interface.state!s}")
        self.connected = True

    def close(self) -> None:
        if self.connected and not self.parent.has_user_supplied_interface:
            self.can_interface.shutdown()
        if self.saved_filters:
            self.can_interface.set_filters(self.saved_filters)
        self.connected = False

    def transmit(self, payload: bytes) -> None:
        frame = Message(
            arbitration_id=self.parent.can_id_master.id,
            is_extended_id=True if self.parent.can_id_master.is_extended else False,
            is_fd=self.parent.fd,
            data=payload,
        )
        self.can_interface.send(frame)

    def read(self) -> Optional[Frame]:
        if not self.connected:
            return None
        try:
            frame = self.can_interface.recv(self.timeout)
        except CanError:
            return None
        else:
            if frame is None or not len(frame.data):
                return None  # Timeout condition.
            if not self.software_filter.accept(frame):
                return None  # Filter out unwanted traffic.
            extended = frame.is_extended_id
            identifier = Identifier.make_identifier(frame.arbitration_id, extended)
            return Frame(
                id_=identifier,
                dlc=frame.dlc,
                data=frame.data,
                timestamp=seconds_to_nanoseconds(frame.timestamp),
            )

    def get_timestamp_resolution(self) -> int:
        return 10 * 1000


class EmptyHeader:
    """There is no header for XCP on CAN"""

    def pack(self, *args, **kwargs):
        return b""


class Can(BaseTransport):
    """"""

    MAX_DATAGRAM_SIZE = 7
    HEADER = EmptyHeader()
    HEADER_SIZE = 0

    def __init__(self, config, policy=None, transport_layer_interface: Optional[BusABC] = None):
        super().__init__(config, policy, transport_layer_interface)
        self.load_config(config)
        self.useDefaultListener = self.config.use_default_listener
        self.can_id_master = Identifier(self.config.can_id_master)
        self.can_id_slave = Identifier(self.config.can_id_slave)

        # Regarding CAN-FD s. AUTOSAR CP Release 4.3.0, Requirements on CAN; [SRS_Can_01160] Padding of bytes due to discrete CAN FD DLC]:
        #   "... If a PDU does not exactly match these configurable sizes the unused bytes shall be padded."
        #
        self.fd = self.config.fd
        self.daq_identifier = []
        if self.config.daq_identifier:
            for daq_id in self.config.daq_identifier:
                self.daq_identifier.append(Identifier(daq_id))
        self.max_dlc_required = self.config.max_dlc_required
        self.padding_value = self.config.padding_value
        if transport_layer_interface is None:
            self.interface_name = self.config.interface
            self.interface_configuration = detect_available_configs(interfaces=[self.interface_name])
            parameters = self.get_interface_parameters()
        else:
            self.interface_name = "custom"
            # print("TRY GET PARAMs", self.get_interface_parameters())
            parameters = {}
        self.can_interface = PythonCanWrapper(self, self.interface_name, config.timeout, **parameters)
        self.logger.info(f"XCPonCAN - Interface-Type: {self.interface_name!r} Parameters: {list(parameters.items())}")
        self.logger.info(
            f"XCPonCAN - Master-ID (Tx): 0x{self.can_id_master.id:08X}{self.can_id_master.type_str} -- "
            f"Slave-ID (Rx): 0x{self.can_id_slave.id:08X}{self.can_id_slave.type_str}"
        )

    def get_interface_parameters(self) -> Dict[str, Any]:
        result = dict(channel=self.config.channel)

        can_interface_config_class = CAN_INTERFACE_MAP[self.interface_name]

        # Optional base class parameters.
        optional_parameters = [(p, p.removeprefix("has_")) for p in can_interface_config_class.OPTIONAL_BASE_PARAMS]
        for o, n in optional_parameters:
            opt = getattr(can_interface_config_class, o)
            value = getattr(self.config, n)
            if opt:
                if value is not None:
                    result[n] = value
            elif value is not None:
                self.logger.warning(f"XCPonCAN - {self.interface_name!r} has no support for parameter {n!r}.")
        # Parameter names that need to be mapped.
        for base_name, name in can_interface_config_class.CAN_PARAM_MAP.items():
            value = getattr(self.config, base_name)
            if name is not None and value is not None:
                result[name] = value
        # Interface specific parameters.
        cxx = getattr(self.config, self.interface_name)
        for name in can_interface_config_class.class_own_traits().keys():
            value = getattr(cxx, name)
            if value is not None:
                result[name] = value
        return result

    def data_received(self, payload: bytes, recv_timestamp: int):
        self.process_response(
            payload,
            len(payload),
            counter=(self.counter_received + 1) & 0xFFFF,
            recv_timestamp=recv_timestamp,
        )

    def listen(self):
        while True:
            if self.closeEvent.is_set():
                return
            frame = self.can_interface.read()
            if frame:
                self.data_received(frame.data, frame.timestamp)

    def connect(self):
        if self.useDefaultListener:
            self.start_listener()
        try:
            self.can_interface.connect()
        except CanInitializationError:
            console.print("[red]\nThere may be a problem with the configuration of your CAN-interface.\n")
            console.print(f"[grey]Current configuration of interface {self.interface_name!r}:")
            console.print(self.interface_configuration)
            raise
        self.status = 1  # connected

    def send(self, frame: bytes) -> None:
        # send the request
        self.pre_send_timestamp = self.timestamp.value
        self.can_interface.transmit(payload=pad_frame(frame, self.max_dlc_required, self.padding_value))
        self.post_send_timestamp = self.timestamp.value

    def close_connection(self):
        if hasattr(self, "can_interface"):
            self.can_interface.close()

    def close(self):
        self.finish_listener()
        self.close_connection()


def set_DLC(length: int):
    """Return DLC value according to CAN-FD.

    :param length: Length value to be mapped to a valid CAN-FD DLC.
                   ( 0 <= length <= 64)
    """

    if length < 0:
        raise ValueError("Non-negative length value required.")
    elif length <= MAX_DLC_CLASSIC:
        return length
    elif length <= 64:
        for dlc in CAN_FD_DLCS:
            if length <= dlc:
                return dlc
    else:
        raise ValueError("DLC could be at most 64.")


def calculate_filter(ids: list):
    """
    :param ids: An iterable (usually list or tuple) containing CAN identifiers.

    :return: Calculated filter and mask.
    :rtype: tuple (int, int)
    """
    any_extended_ids = any(is_extended_identifier(i) for i in ids)
    raw_ids = [stripIdentifier(i) for i in ids]
    cfilter = functools.reduce(operator.and_, raw_ids)
    cmask = functools.reduce(operator.or_, raw_ids) ^ cfilter
    cmask ^= 0x1FFFFFFF if any_extended_ids else 0x7FF
    return (cfilter, cmask)


class CanInterfaceBase(ABC):
    """
    Base class for custom CAN interfaces.
    This is basically a subset of python-CANs `BusABC`.
    """

    @abstractmethod
    def set_filters(self, filters: Optional[List[Dict[str, Union[int, bool]]]] = None) -> None:
        """Apply filtering to all messages received by this Bus.

        filters:
            A list of dictionaries, each containing a 'can_id', 'can_mask', and 'extended' field, e.g.:
            [{"can_id": 0x11, "can_mask": 0x21, "extended": False}]
        """

    @abstractmethod
    def recv(self, timeout: Optional[float] = None) -> Optional[Message]:
        """Block waiting for a message from the Bus."""

    @abstractmethod
    def send(self, msg: Message) -> None:
        """Transmit a message to the CAN bus."""

    @property
    @abstractmethod
    def filters(self) -> Optional[List[Dict[str, Union[int, bool]]]]:
        """Modify the filters of this bus."""

    @property
    @abstractmethod
    def state(self) -> BusState:
        """Return the current state of the hardware."""

    def __repr__(self):
        return f"{self.__class__.__name__}"



import struct
from collections import deque
from dataclasses import dataclass
from typing import Optional

import serial

import pyxcp.types as types
from pyxcp.transport.base import BaseTransport


@dataclass
class HeaderValues:
    length: int = 0
    counter: int = 0
    filler: int = 0


RECV_SIZE = 16384


class SxI(BaseTransport):
    """"""

    def __init__(self, config=None, policy=None, transport_layer_interface: Optional[serial.Serial] = None) -> None:
        super().__init__(config, policy, transport_layer_interface)
        self.load_config(config)
        self.port_name = self.config.port
        self.baudrate = self.config.bitrate
        self.bytesize = self.config.bytesize
        self.parity = self.config.parity
        self.stopbits = self.config.stopbits
        self.mode = self.config.mode
        self.header_format = self.config.header_format
        self.tail_format = self.config.tail_format
        self.framing = self.config.framing
        self.esc_sync = self.config.esc_sync
        self.esc_esc = self.config.esc_esc
        self.make_header()
        self.comm_port: serial.Serial

        if self.has_user_supplied_interface and transport_layer_interface:
            self.comm_port = transport_layer_interface
        else:
            self.logger.info(f"XCPonSxI - trying to open serial comm_port {self.port_name!r}.")
            try:
                self.comm_port = serial.Serial(
                    port=self.port_name,
                    baudrate=self.baudrate,
                    bytesize=self.bytesize,
                    parity=self.parity,
                    stopbits=self.stopbits,
                    timeout=self.timeout,
                    write_timeout=self.timeout,
                )
            except serial.SerialException as e:
                self.logger.critical(f"XCPonSxI - {e}")
                raise
        self._packets = deque()

    def __del__(self) -> None:
        self.close_connection()

    def make_header(self) -> None:
        def unpack_len(args):
            (length,) = args
            return HeaderValues(length=length)

        def unpack_len_counter(args):
            length, counter = args
            return HeaderValues(length=length, counter=counter)

        def unpack_len_filler(args):
            length, filler = args
            return HeaderValues(length=length, filler=filler)

        HEADER_FORMATS = {
            "HEADER_LEN_BYTE": ("B", unpack_len),
            "HEADER_LEN_CTR_BYTE": ("BB", unpack_len_counter),
            "HEADER_LEN_FILL_BYTE": ("BB", unpack_len_filler),
            "HEADER_LEN_WORD": ("H", unpack_len),
            "HEADER_LEN_CTR_WORD": ("HH", unpack_len_counter),
            "HEADER_LEN_FILL_WORD": ("HH", unpack_len_filler),
        }
        fmt, unpacker = HEADER_FORMATS[self.header_format]
        self.HEADER = struct.Struct(f"<{fmt}")
        self.HEADER_SIZE = self.HEADER.size
        self.unpacker = unpacker

    def connect(self) -> None:
        self.logger.info(
            f"XCPonSxI - serial comm_port {self.comm_port.portstr!r} openend [{self.baudrate}/{self.bytesize}-{self.parity}-{self.stopbits}]"
        )
        self.start_listener()

    def output(self, enable) -> None:
        if enable:
            self.comm_port.rts = False
            self.comm_port.dtr = False
        else:
            self.comm_port.rts = True
            self.comm_port.dtr = True

    def flush(self) -> None:
        self.comm_port.flush()

    def start_listener(self) -> None:
        super().start_listener()

    def listen(self) -> None:
        while True:
            if self.closeEvent.is_set():
                return
            if not self.comm_port.in_waiting:
                continue

            recv_timestamp = self.timestamp.value
            header_values = self.unpacker(self.HEADER.unpack(self.comm_port.read(self.HEADER_SIZE)))
            length, counter, _ = header_values.length, header_values.counter, header_values.filler

            response = self.comm_port.read(length)
            self.timing.stop()

            if len(response) != length:
                raise types.FrameSizeError("Size mismatch.")
            self.process_response(response, length, counter, recv_timestamp)

    def send(self, frame) -> None:
        self.pre_send_timestamp = self.timestamp.value
        self.comm_port.write(frame)
        self.post_send_timestamp = self.timestamp.value

    def close_connection(self) -> None:
        if hasattr(self, "comm_port") and self.comm_port.is_open and not self.has_user_supplied_interface:
            self.comm_port.close()



#!/usr/bin/env python
"""Implements error-handling according to XCP spec.
"""
import functools
import logging
import threading
import time
import types
from collections import namedtuple
from typing import Generic, List, Optional, TypeVar

import can

from pyxcp.errormatrix import ERROR_MATRIX, Action, PreAction
from pyxcp.types import COMMAND_CATEGORIES, XcpError, XcpResponseError, XcpTimeoutError


handle_errors = True  # enable/disable XCP error-handling.


class SingletonBase:
    _lock = threading.Lock()

    def __new__(cls, *args, **kws):
        # Double-Checked Locking
        if not hasattr(cls, "_instance"):
            try:
                cls._lock.acquire()
                if not hasattr(cls, "_instance"):
                    cls._instance = super().__new__(cls)
            finally:
                cls._lock.release()
        return cls._instance


Function = namedtuple("Function", "fun arguments")  # store: var | load: var


class InternalError(Exception):
    """Indicates an internal error, like invalid service."""


class SystemExit(Exception):
    """"""

    def __init__(self, msg: str, error_code: int = None, *args, **kws):
        super().__init__(*args, **kws)
        self.error_code = error_code
        self.msg = msg

    def __str__(self):
        return f"SystemExit(error_code={self.error_code}, message={self.msg!r})"

    __repr__ = __str__


class UnrecoverableError(Exception):
    """"""


def func_name(func):
    return func.__qualname__ if func is not None else None


def getErrorHandler(service):
    """"""
    return ERROR_MATRIX.get(service)


def getTimeoutHandler(service):
    """"""
    handler = getErrorHandler(service)
    if handler is None:
        raise InternalError("Invalid Service")
    return handler.get(XcpError.ERR_TIMEOUT)


def getActions(service, error_code):
    """"""
    error_str = str(error_code)
    if error_code == XcpError.ERR_TIMEOUT:
        preActions, actions = getTimeoutHandler(service)
    else:
        eh = getErrorHandler(service)
        if eh is None:
            raise InternalError(f"Invalid Service 0x{service:02x}")
        # print(f"Try to handle error -- Service: {service.name} Error-Code: {error_code}")
        handler = eh.get(error_str)
        if handler is None:
            raise SystemExit(f"Service {service.name!r} has no handler for {error_code}.", error_code=error_code)
        preActions, actions = handler
    return preActions, actions


def actionIter(actions):
    """Iterate over action from :file:`errormatrix.py`"""
    if isinstance(actions, (tuple, list)):
        yield from actions
    else:
        yield actions


class Arguments:
    """Container for positional and keyword arguments.

    Parameters
    ----------
        args: tuple
            Positional arguments
        kwargs: dict
            Keyword arguments.
    """

    def __init__(self, args=None, kwargs=None):
        if args is None:
            self.args = ()
        else:
            if not hasattr(args, "__iter__"):
                self.args = (args,)
            else:
                self.args = tuple(args)
        self.kwargs = kwargs or {}

    def __str__(self) -> str:
        res = f"{self.__class__.__name__}(ARGS = {self.args}, KWS = {self.kwargs})"
        return res

    def __eq__(self, other) -> bool:
        return (self.args == other.args if other is not None else False) and (
            self.kwargs == other.kwargs if other is not None else False
        )

    __repr__ = __str__


class Repeater:
    """A required action of some XCP errorhandler is repetition.

    Parameters
    ----------
        initial_value: int
            The actual values are predetermined by XCP:
                - REPEAT (one time)
                - REPEAT_2_TIMES (two times)
                - REPEAT_INF_TIMES ("forever")
    """

    REPEAT = 1
    REPEAT_2_TIMES = 2
    INFINITE = -1

    def __init__(self, initial_value: int):
        self._counter = initial_value
        # print("\tREPEATER ctor", hex(id(self)))

    def repeat(self):
        """Check if repetition is required.

        Returns
        -------
            bool
        """
        # print("\t\tCOUNTER:", hex(id(self)), self._counter)
        if self._counter == Repeater.INFINITE:
            return True
        elif self._counter > 0:
            self._counter -= 1
            return True
        else:
            return False


def display_error():
    """Display error information.

    TODO: callback.

    """


class Handler:
    """"""

    def __init__(self, instance, func, arguments, error_code=None):
        self.instance = instance
        if hasattr(func, "__closure__") and func.__closure__:
            self.func = func.__closure__[0].cell_contents  # Use original, undecorated function to prevent
            # nasty recursion problems.
        else:
            self.func = func
        self.arguments = arguments
        self.service = self.instance.service
        self._error_code: int = 0
        if error_code is not None:
            self._error_code = error_code
        self._repeater = None
        self.logger = logging.getLogger("PyXCP")

    def __str__(self):
        return f"Handler(func = {func_name(self.func)} -- {self.arguments} service = {self.service} error_code = {self.error_code})"

    def __eq__(self, other):
        if other is None:
            return False
        return (self.instance == other.instance) and (self.func == other.func) and (self.arguments == other.arguments)

    @property
    def error_code(self) -> int:
        return self._error_code

    @error_code.setter
    def error_code(self, value: int) -> None:
        self._error_code = value

    @property
    def repeater(self):
        # print("\tGet repeater", hex(id(self._repeater)), self._repeater is None)
        return self._repeater

    @repeater.setter
    def repeater(self, value):
        # print("\tSet repeater", hex(id(value)))
        self._repeater = value

    def execute(self):
        self.logger.debug(f"Execute({func_name(self.func)} -- {self.arguments})")
        if isinstance(self.func, types.MethodType):
            return self.func(*self.arguments.args, **self.arguments.kwargs)
        else:
            return self.func(self.instance, *self.arguments.args, **self.arguments.kwargs)

    def actions(self, preActions, actions):
        """Preprocess errorhandling pre-actions and actions."""
        result_pre_actions = []
        result_actions = []
        repetitionCount = 0
        for item in actionIter(preActions):
            if item == PreAction.NONE:
                pass
            elif item == PreAction.WAIT_T7:
                time.sleep(0.02)  # Completely arbitrary for now.
            elif item == PreAction.SYNCH:
                fn = Function(self.instance.synch, Arguments())
                result_pre_actions.append(fn)
            elif item == PreAction.GET_SEED_UNLOCK:
                raise NotImplementedError("Pre-action GET_SEED_UNLOCK")
            elif item == PreAction.SET_MTA:
                fn = Function(self.instance.setMta, Arguments(self.instance.mta))
                result_pre_actions.append(fn)
            elif item == PreAction.SET_DAQ_PTR:
                fn = Function(self.instance.setDaqPtr, Arguments(self.instance.currentDaqPtr))
            elif item == PreAction.START_STOP_X:
                raise NotImplementedError("Pre-action START_STOP_X")
            elif item == PreAction.REINIT_DAQ:
                raise NotImplementedError("Pre-action REINIT_DAQ")
            elif item == PreAction.DISPLAY_ERROR:
                pass
            elif item == PreAction.DOWNLOAD:
                raise NotImplementedError("Pre-action DOWNLOAD")
            elif item == PreAction.PROGRAM:
                raise NotImplementedError("Pre-action PROGRAM")
            elif item == PreAction.UPLOAD:
                raise NotImplementedError("Pre-action UPLOAD")
            elif item == PreAction.UNLOCK_SLAVE:
                resource = COMMAND_CATEGORIES.get(self.instance.service)  # noqa: F841
                raise NotImplementedError("Pre-action UNLOCK_SLAVE")
        for item in actionIter(actions):
            if item == Action.NONE:
                pass
            elif item == Action.DISPLAY_ERROR:
                raise SystemExit("Could not proceed due to unhandled error (DISPLAY_ERROR).", self.error_code)
            elif item == Action.RETRY_SYNTAX:
                raise SystemExit("Could not proceed due to unhandled error (RETRY_SYNTAX).", self.error_code)
            elif item == Action.RETRY_PARAM:
                raise SystemExit("Could not proceed due to unhandled error (RETRY_PARAM).", self.error_code)
            elif item == Action.USE_A2L:
                raise SystemExit("Could not proceed due to unhandled error (USE_A2L).", self.error_code)
            elif item == Action.USE_ALTERATIVE:
                raise SystemExit(
                    "Could not proceed due to unhandled error (USE_ALTERATIVE).", self.error_code
                )  # TODO: check alternatives.
            elif item == Action.REPEAT:
                repetitionCount = Repeater.REPEAT
            elif item == Action.REPEAT_2_TIMES:
                repetitionCount = Repeater.REPEAT_2_TIMES
            elif item == Action.REPEAT_INF_TIMES:
                repetitionCount = Repeater.INFINITE
            elif item == Action.RESTART_SESSION:
                raise SystemExit("Could not proceed due to unhandled error (RESTART_SESSION).", self.error_code)
            elif item == Action.TERMINATE_SESSION:
                raise SystemExit("Could not proceed due to unhandled error (TERMINATE_SESSION).", self.error_code)
            elif item == Action.SKIP:
                pass
            elif item == Action.NEW_FLASH_WARE:
                raise SystemExit("Could not proceed due to unhandled error (NEW_FLASH_WARE)", self.error_code)
        return result_pre_actions, result_actions, Repeater(repetitionCount)


T = TypeVar("T")


class HandlerStack(Generic[T]):
    """"""

    def __init__(self) -> None:
        self._stack: List[T] = []

    def push(self, value: T):
        if value != self.tos():
            self._stack.append(value)

    def pop(self) -> None:
        if len(self) > 0:
            self._stack.pop()

    def tos(self) -> Optional[T]:
        if len(self) > 0:
            return self._stack[-1]
        else:
            return None
            # raise ValueError("empty stack.")

    def empty(self) -> bool:
        return self._stack == []

    def __len__(self) -> int:
        return len(self._stack)

    def __repr__(self) -> str:
        result = []
        for idx in range(len(self)):
            result.append(str(self[idx]))
        return "\n".join(result)

    def __getitem__(self, ndx: int) -> T:
        return self._stack[ndx]

    __str__ = __repr__


class Executor(SingletonBase):
    """"""

    def __init__(self):
        self.handlerStack = HandlerStack()
        self.repeater = None
        self.logger = logging.getLogger("PyXCP")
        self.previous_error_code = None
        self.error_code = None
        self.func = None
        self.arguments = None

    def __call__(self, inst, func, arguments):
        self.inst = inst
        self.func = func
        self.arguments = arguments
        handler = Handler(inst, func, arguments)
        self.handlerStack.push(handler)
        connect_retries = inst.config.connect_retries
        try:
            while True:
                try:
                    handler = self.handlerStack.tos()
                    res = handler.execute()
                except XcpResponseError as e:
                    # self.logger.critical(f"XcpResponseError [{e.get_error_code()}]")
                    self.error_code = e.get_error_code()
                    handler.error_code = self.error_code
                except XcpTimeoutError:
                    is_connect = func.__name__ == "connect"
                    self.logger.warning(f"XcpTimeoutError -- Service: {func.__name__!r}")
                    self.error_code = XcpError.ERR_TIMEOUT
                    handler.error_code = self.error_code
                    if is_connect and connect_retries is not None:
                        if connect_retries == 0:
                            raise XcpTimeoutError("Maximum CONNECT retries reached.")
                        connect_retries -= 1
                except TimeoutError:
                    raise
                except can.CanError:
                    # self.logger.critical(f"Exception raised by Python CAN [{str(e)}]")
                    raise
                except Exception:
                    # self.logger.critical(f"Exception [{str(e)}]")
                    raise
                else:
                    self.error_code = None
                    self.handlerStack.pop()
                    if self.handlerStack.empty():
                        return res

                if self.error_code == XcpError.ERR_CMD_SYNCH:
                    # Don't care about SYNCH for now...
                    self.inst.logger.info("SYNCH received.")
                    continue

                if self.error_code is not None:
                    preActions, actions, repeater = handler.actions(*getActions(inst.service, self.error_code))
                    if handler.repeater is None:
                        handler.repeater = repeater
                    for f, a in reversed(preActions):
                        self.handlerStack.push(Handler(inst, f, a, self.error_code))
                self.previous_error_code = self.error_code
                if handler.repeater:
                    if handler.repeater.repeat():
                        continue
                    else:
                        raise UnrecoverableError(
                            f"Max. repetition count reached while trying to execute service {handler.func.__name__!r}."
                        )
        finally:
            # cleanup of class variables
            self.previous_error_code = None
            while not self.handlerStack.empty():
                self.handlerStack.pop()
            self.error_code = None
            self.func = None
            self.arguments = None


def disable_error_handling(value: bool):
    """Disable XCP error-handling (mainly for performance reasons)."""

    global handle_errors
    handle_errors = not bool(value)


def wrapped(func):
    """This decorator is XCP error-handling enabled."""

    @functools.wraps(func)
    def inner(*args, **kwargs):
        if handle_errors:
            inst = args[0]  # First parameter is 'self'.
            arguments = Arguments(args[1:], kwargs)
            executor = Executor()
            res = executor(inst, func, arguments)
        else:
            res = func(*args, **kwargs)
        return res

    return inner



#!/usr/bin/env python

import pytest

from pyxcp.cpp_ext.cpp_ext import Bin, DaqList, McObject  # # noqa:  F401
from pyxcp.daq_stim.optimize import make_continuous_blocks
from pyxcp.daq_stim.optimize.binpacking import first_fit_decreasing


DAQ_LISTS_1 = [
    DaqList(
        "test",
        0,
        False,
        False,
        [
            ("voltage1", 0x0080051A - 4, 0, "F32"),  # 80053a
            ("voltage2", 0x0080051E - 4, 0, "F32"),
            ("voltage3", 0x00800522 - 4, 0, "F32"),
            ("voltage4", 0x00800526 - 4, 0, "F32"),
            ("sine_wave", 0x00800512, 0, "F32"),
        ],
    ),
]

DAQ_LISTS_2 = [
    DaqList(
        "test",
        0,
        False,
        False,
        [
            ("voltage1", 0x0080051A, 0, "F32"),  # 80053a
            ("voltage2", 0x0080051E, 0, "F32"),
            ("voltage3", 0x00800522, 0, "F32"),
            ("voltage4", 0x00800526, 0, "F32"),
            ("sq0", 0x00800516, 0, "U8"),
            ("sq1", 0x00800517, 0, "U8"),
            ("sine_wave", 0x00800512, 0, "F32"),
        ],
    ),
]

max_payload_size = 64
max_payload_size_first = 64

EXPECTED_BLOCK_VAR_1_64 = [
    McObject(
        name="",
        address=8389906,
        ext=0,
        data_type="",
        length=20,
        components=[
            McObject(name="sine_wave", address=8389906, ext=0, data_type="F32", length=4, components=[]),
            McObject(name="voltage1", address=8389910, ext=0, data_type="F32", length=4, components=[]),
            McObject(name="voltage2", address=8389914, ext=0, data_type="F32", length=4, components=[]),
            McObject(name="voltage3", address=8389918, ext=0, data_type="F32", length=4, components=[]),
            McObject(name="voltage4", address=8389922, ext=0, data_type="F32", length=4, components=[]),
        ],
    )
]

EXPECTED_BLOCK_VAR_1_8 = [
    McObject(
        name="",
        address=8389906,
        ext=0,
        data_type="",
        length=4,
        components=[McObject(name="sine_wave", address=8389906, ext=0, data_type="F32", length=4, components=[])],
    ),
    McObject(
        name="",
        address=8389910,
        ext=0,
        data_type="",
        length=4,
        components=[McObject(name="voltage1", address=8389910, ext=0, data_type="F32", length=4, components=[])],
    ),
    McObject(
        name="",
        address=8389914,
        ext=0,
        data_type="",
        length=4,
        components=[McObject(name="voltage2", address=8389914, ext=0, data_type="F32", length=4, components=[])],
    ),
    McObject(
        name="",
        address=8389918,
        ext=0,
        data_type="",
        length=4,
        components=[McObject(name="voltage3", address=8389918, ext=0, data_type="F32", length=4, components=[])],
    ),
    McObject(
        name="",
        address=8389922,
        ext=0,
        data_type="",
        length=4,
        components=[McObject(name="voltage4", address=8389922, ext=0, data_type="F32", length=4, components=[])],
    ),
]

EXPECTED_BLOCK_VAR_2_64 = [
    McObject(
        name="",
        address=8389906,
        ext=0,
        data_type="",
        length=6,
        components=[
            McObject(name="sine_wave", address=8389906, ext=0, data_type="F32", length=4, components=[]),
            McObject(name="sq0", address=8389910, ext=0, data_type="U8", length=1, components=[]),
            McObject(name="sq1", address=8389911, ext=0, data_type="U8", length=1, components=[]),
        ],
    ),
    McObject(
        name="",
        address=8389914,
        ext=0,
        data_type="",
        length=16,
        components=[
            McObject(name="voltage1", address=8389914, ext=0, data_type="F32", length=4, components=[]),
            McObject(name="voltage2", address=8389918, ext=0, data_type="F32", length=4, components=[]),
            McObject(name="voltage3", address=8389922, ext=0, data_type="F32", length=4, components=[]),
            McObject(name="voltage4", address=8389926, ext=0, data_type="F32", length=4, components=[]),
        ],
    ),
]

EXPECTED_BLOCK_VAR_2_8 = [
    McObject(
        name="",
        address=8389906,
        ext=0,
        data_type="",
        length=6,
        components=[
            McObject(name="sine_wave", address=8389906, ext=0, data_type="F32", length=4, components=[]),
            McObject(name="sq0", address=8389910, ext=0, data_type="U8", length=1, components=[]),
            McObject(name="sq1", address=8389911, ext=0, data_type="U8", length=1, components=[]),
        ],
    ),
    McObject(
        name="",
        address=8389914,
        ext=0,
        data_type="",
        length=4,
        components=[
            McObject(name="voltage1", address=8389914, ext=0, data_type="F32", length=4, components=[]),
        ],
    ),
    McObject(
        name="",
        address=8389918,
        ext=0,
        data_type="",
        length=4,
        components=[
            McObject(name="voltage2", address=8389918, ext=0, data_type="F32", length=4, components=[]),
        ],
    ),
    McObject(
        name="",
        address=8389922,
        ext=0,
        data_type="",
        length=4,
        components=[
            McObject(name="voltage3", address=8389922, ext=0, data_type="F32", length=4, components=[]),
        ],
    ),
    McObject(
        name="",
        address=8389926,
        ext=0,
        data_type="",
        length=4,
        components=[
            McObject(name="voltage4", address=8389926, ext=0, data_type="F32", length=4, components=[]),
        ],
    ),
]

BIN_PACKING_VAR1_8 = (
    (
        7,
        3,
        [
            McObject(
                name="",
                address=8389906,
                ext=0,
                data_type="",
                length=4,
                components=[
                    McObject(name="sine_wave", address=8389906, ext=0, data_type="F32", length=4, components=[]),
                ],
            )
        ],
    ),
    (
        7,
        3,
        [
            McObject(
                name="",
                address=8389910,
                ext=0,
                data_type="",
                length=4,
                components=[
                    McObject(name="voltage1", address=8389910, ext=0, data_type="F32", length=4, components=[]),
                ],
            )
        ],
    ),
    (
        7,
        3,
        [
            McObject(
                name="",
                address=8389914,
                ext=0,
                data_type="",
                length=4,
                components=[
                    McObject(name="voltage2", address=8389914, ext=0, data_type="F32", length=4, components=[]),
                ],
            )
        ],
    ),
    (
        7,
        3,
        [
            McObject(
                name="",
                address=8389918,
                ext=0,
                data_type="",
                length=4,
                components=[
                    McObject(name="voltage3", address=8389918, ext=0, data_type="F32", length=4, components=[]),
                ],
            )
        ],
    ),
)

BIN_PACKING_VAR2_8 = (
    (
        7,
        1,
        [
            McObject(
                name="",
                address=8389906,
                ext=0,
                data_type="",
                length=6,
                components=[
                    McObject(name="sine_wave", address=8389906, ext=0, data_type="F32", length=4, components=[]),
                    McObject(name="sq0", address=8389910, ext=0, data_type="U8", length=1, components=[]),
                    McObject(name="sq1", address=8389911, ext=0, data_type="U8", length=1, components=[]),
                ],
            )
        ],
    ),
    (
        7,
        3,
        [
            McObject(
                name="",
                address=8389914,
                ext=0,
                data_type="",
                length=4,
                components=[
                    McObject(name="voltage1", address=8389914, ext=0, data_type="F32", length=4, components=[]),
                ],
            )
        ],
    ),
    (
        7,
        3,
        [
            McObject(
                name="",
                address=8389918,
                ext=0,
                data_type="",
                length=4,
                components=[
                    McObject(name="voltage2", address=8389918, ext=0, data_type="F32", length=4, components=[]),
                ],
            )
        ],
    ),
    (
        7,
        3,
        [
            McObject(
                name="",
                address=8389922,
                ext=0,
                data_type="",
                length=4,
                components=[
                    McObject(name="voltage3", address=8389922, ext=0, data_type="F32", length=4, components=[]),
                ],
            )
        ],
    ),
    (
        7,
        3,
        [
            McObject(
                name="",
                address=8389926,
                ext=0,
                data_type="",
                length=4,
                components=[
                    McObject(name="voltage4", address=8389926, ext=0, data_type="F32", length=4, components=[]),
                ],
            )
        ],
    ),
)

BIN_PACKING_VAR1_64 = (
    (
        63,
        43,
        [
            McObject(
                name="",
                address=8389906,
                ext=0,
                data_type="",
                length=20,
                components=[
                    McObject(name="sine_wave", address=8389906, ext=0, data_type="F32", length=4, components=[]),
                    McObject(name="voltage1", address=8389910, ext=0, data_type="F32", length=4, components=[]),
                    McObject(name="voltage2", address=8389914, ext=0, data_type="F32", length=4, components=[]),
                    McObject(name="voltage3", address=8389918, ext=0, data_type="F32", length=4, components=[]),
                    McObject(name="voltage4", address=8389922, ext=0, data_type="F32", length=4, components=[]),
                ],
            )
        ],
    ),
)

BIN_PACKING_VAR2_64 = (
    63,
    41,
    [
        McObject(
            name="",
            address=8389914,
            ext=0,
            data_type="",
            length=16,
            components=[
                McObject(name="voltage1", address=8389914, ext=0, data_type="F32", length=4, components=[]),
                McObject(name="voltage2", address=8389918, ext=0, data_type="F32", length=4, components=[]),
                McObject(name="voltage3", address=8389922, ext=0, data_type="F32", length=4, components=[]),
                McObject(name="voltage4", address=8389926, ext=0, data_type="F32", length=4, components=[]),
            ],
        ),
        McObject(
            name="",
            address=8389906,
            ext=0,
            data_type="",
            length=6,
            components=[
                McObject(name="sine_wave", address=8389906, ext=0, data_type="F32", length=4, components=[]),
                McObject(name="sq0", address=8389910, ext=0, data_type="U8", length=1, components=[]),
                McObject(name="sq1", address=8389911, ext=0, data_type="U8", length=1, components=[]),
            ],
        ),
    ],
)


@pytest.mark.parametrize(
    "daq_lists, expected_blocks, payload_size",
    [
        (DAQ_LISTS_1, EXPECTED_BLOCK_VAR_1_64, 64),
        (DAQ_LISTS_1, EXPECTED_BLOCK_VAR_1_8, 8),
        (DAQ_LISTS_2, EXPECTED_BLOCK_VAR_2_64, 64),
        (DAQ_LISTS_2, EXPECTED_BLOCK_VAR_2_8, 8),
    ],
)
def test_make_continuous_blocks(daq_lists, expected_blocks, payload_size):
    daq_list = daq_lists[0]
    blocks = make_continuous_blocks(daq_list.measurements, payload_size - 1, payload_size - 1)
    for block, expected_block in zip(blocks, expected_blocks):
        assert block == expected_block


@pytest.mark.parametrize(
    "blocks, expected_blocks, payload_size",
    [
        (EXPECTED_BLOCK_VAR_1_8, BIN_PACKING_VAR1_8, 8),
        (EXPECTED_BLOCK_VAR_2_8, BIN_PACKING_VAR2_8, 8),
        (EXPECTED_BLOCK_VAR_1_64, BIN_PACKING_VAR1_64, 64),
        (EXPECTED_BLOCK_VAR_2_64, BIN_PACKING_VAR2_64, 64),
    ],
)
def test_first_fit_decreasing(blocks, expected_blocks, payload_size):
    res = first_fit_decreasing(blocks, payload_size - 1, payload_size - 1)  # noqa:  F841
    # for entry in res:
    #    print(entry.size, entry.residual_capacity, entry.entries)



#!/usr/bin/env python
"""
Parse (transport-layer specific) command line parameters
and create a XCP master instance.
"""

import warnings
from dataclasses import dataclass
from typing import List

from pyxcp.config import (  # noqa: F401
    create_application,
    get_application,
    reset_application,
)
from pyxcp.master import Master


warnings.simplefilter("always")


@dataclass
class Option:
    short_opt: str
    long_opt: str = ""
    dest: str = ""
    help: str = ""
    type: str = ""
    default: str = ""
    action: str = ""


class FakeParser:
    """Parser that collects arguments for later processing."""

    def __init__(self):
        self.options = []

    def add_argument(self, short_opt, long_opt="", dest="", help="", type=None, default=None, action=None):
        """Collect argument definitions without issuing warnings."""
        self.options.append(Option(short_opt, long_opt, dest, help, type, default, action))


class ArgumentParser:
    """Argument parser for pyXCP applications.

    This class provides a way to add custom command-line arguments to pyXCP applications.
    It also supports a callout function that will be called with the master instance and
    the parsed arguments.
    """

    def __init__(self, callout=None, description=None, *args, **kws):
        self._parser = FakeParser()
        self._callout = callout
        self._description = description

    def run(self, policy=None, transport_layer_interface=None):
        """Create and configure a master instance.

        Args:
            policy: Optional policy to use for the master
            transport_layer_interface: Optional transport layer interface to use

        Returns:
            A configured master instance
        """
        # Create the application with custom arguments and callout
        application = get_application(self.parser.options, self._callout)

        # Create the master instance
        master = Master(
            application.transport.layer, config=application, policy=policy, transport_layer_interface=transport_layer_interface
        )

        # If there's a callout function, call it with the master and args
        if application.callout is not None and hasattr(application, "custom_args"):
            args = application.custom_args.get_args()
            application.callout(master, args)

        return master

    @property
    def parser(self):
        return self._parser



#!/usr/bin/env python
import datetime
import functools
import operator
import sys
from binascii import hexlify
from enum import IntEnum
from time import perf_counter, sleep
from typing import Any, List, Optional, Union

import chardet
import pytz

from pyxcp.cpp_ext.cpp_ext import TimestampInfo


def hexDump(arr: Union[bytes | bytearray]):
    if isinstance(arr, (bytes, bytearray)):
        size = len(arr)
        try:
            arr = arr.hex()
        except BaseException:  # noqa: B036
            arr = hexlify(arr).decode("ascii")
        return "[{}]".format(" ".join([arr[i * 2 : (i + 1) * 2] for i in range(size)]))
    elif isinstance(arr, (list, tuple)):
        arr = bytes(arr)
        size = len(arr)
        try:
            arr = arr.hex()
        except BaseException:  # noqa: B036
            arr = hexlify(arr).decode("ascii")
        return "[{}]".format(" ".join([arr[i * 2 : (i + 1) * 2] for i in range(size)]))
    else:
        return "[{}]".format(" ".join([f"{x:02x}" for x in arr]))


def seconds_to_nanoseconds(value: float) -> int:
    return int(value * 1_000_000_000)


def slicer(iterable, sliceLength, converter=None):
    if converter is None:
        converter = type(iterable)
    length = len(iterable)
    return [converter(iterable[item : item + sliceLength]) for item in range(0, length, sliceLength)]


def functools_reduce_iconcat(a):
    return functools.reduce(operator.iconcat, a, [])


def flatten(*args):
    """Flatten a list of lists into a single list.

    s. https://stackoverflow.com/questions/952914/how-do-i-make-a-flat-list-out-of-a-list-of-lists
    """
    return functools.reduce(operator.iconcat, args, [])


def getPythonVersion():
    return sys.version_info


def decode_bytes(byte_str: bytes) -> str:
    """Decode bytes with the help of chardet"""
    encoding = chardet.detect(byte_str).get("encoding")
    if not encoding:
        return byte_str.decode("ascii", "ignore")
    else:
        return byte_str.decode(encoding)


PYTHON_VERSION = getPythonVersion()


def short_sleep():
    sleep(0.0005)


def delay(amount: float):
    """Performe a busy-wait delay, which is much more precise than `time.sleep`"""

    start = perf_counter()
    while perf_counter() < start + amount:
        pass


class CurrentDatetime(TimestampInfo):

    def __init__(self, timestamp_ns: int):
        TimestampInfo.__init__(self, timestamp_ns)
        timezone = pytz.timezone(self.timezone)
        dt = datetime.datetime.fromtimestamp(timestamp_ns / 1_000_000_000.0)
        self.utc_offset = int(timezone.utcoffset(dt).total_seconds() / 60)
        self.dst_offset = int(timezone.dst(dt).total_seconds() / 60)

    def __str__(self):
        return f"""CurrentDatetime(
    datetime="{datetime.datetime.fromtimestamp(self.timestamp_ns / 1_000_000_000.0)!s}",
    timezone="{self.timezone}",
    timestamp_ns={self.timestamp_ns},
    utc_offset={self.utc_offset},
    dst_offset={self.dst_offset}
)"""


def enum_from_str(enum_class: IntEnum, enumerator: str) -> IntEnum:
    """Create an `IntEnum` instance from an enumerator `str`.

    Parameters
    ----------
    enum_class: IntEnum

    enumerator: str

    Example
    -------

    class Color(enum.IntEnum):
        RED = 0
        GREEN = 1
        BLUE = 2

    color: Color = enum_from_str(Color, "GREEN")


    """
    return enum_class(enum_class.__members__.get(enumerator))



#!/usr/bin/env python
"""Lowlevel API reflecting available XCP services.

.. note:: For technical reasons the API is split into two parts;
          common methods (this file) and a Python version specific part.

.. [1] XCP Specification, Part 2 - Protocol Layer Specification
"""
import functools
import logging
import struct
import traceback
import warnings
from typing import Any, Callable, Collection, Dict, List, Optional, Tuple

from pyxcp import checksum, types
from pyxcp.constants import (
    makeBytePacker,
    makeByteUnpacker,
    makeDLongPacker,
    makeDLongUnpacker,
    makeDWordPacker,
    makeDWordUnpacker,
    makeWordPacker,
    makeWordUnpacker,
)
from pyxcp.daq_stim.stim import DaqEventInfo, Stim
from pyxcp.master.errorhandler import SystemExit, disable_error_handling, wrapped
from pyxcp.transport.base import create_transport
from pyxcp.utils import decode_bytes, delay, short_sleep


def broadcasted(func: Callable):
    """"""
    return func


class SlaveProperties(dict):
    """Container class for fixed parameters, like byte-order, maxCTO, ..."""

    def __init__(self, *args, **kws):
        super().__init__(*args, **kws)

    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self = state  # noqa: F841


class Master:
    """Common part of lowlevel XCP API.

    Parameters
    ----------
    transport_name : str
        XCP transport layer name ['can', 'eth', 'sxi']
    config: dict
    """

    def __init__(self, transport_name: Optional[str], config, policy=None, transport_layer_interface=None):
        if transport_name is None:
            raise ValueError("No transport-layer selected")  # Never reached -- to keep type-checkers happy.
        self.ctr = 0
        self.succeeded = True
        self.config = config.general
        self.logger = logging.getLogger("PyXCP")
        disable_error_handling(self.config.disable_error_handling)
        self.transport_name = transport_name.lower()
        transport_config = config.transport
        self.transport = create_transport(transport_name, transport_config, policy, transport_layer_interface)
        self.stim = Stim(self.config.stim_support)
        self.stim.clear()
        self.stim.set_policy_feeder(self.transport.policy.feed)
        self.stim.set_frame_sender(self.transport.block_request)

        # In some cases the transport-layer needs to communicate with us.
        self.transport.parent = self
        self.service = None

        # Policies may issue XCP commands on there own.
        self.transport.policy.xcp_master = self

        # (D)Word (un-)packers are byte-order dependent
        # -- byte-order is returned by CONNECT_Resp (COMM_MODE_BASIC)
        self.BYTE_pack = None
        self.BYTE_unpack = None
        self.WORD_pack = None
        self.WORD_unpack = None
        self.DWORD_pack = None
        self.DWORD_unpack = None
        self.DLONG_pack = None
        self.DLONG_unpack = None
        self.AG_pack = None
        self.AG_unpack = None
        # self.connected = False
        self.mta = types.MtaType(None, None)
        self.currentDaqPtr = None
        self.currentProtectionStatus = None
        self.seed_n_key_dll = self.config.seed_n_key_dll
        self.seed_n_key_function = self.config.seed_n_key_function
        self.seed_n_key_dll_same_bit_width = self.config.seed_n_key_dll_same_bit_width
        self.disconnect_response_optional = self.config.disconnect_response_optional
        self.slaveProperties = SlaveProperties()
        self.slaveProperties.pgmProcessor = SlaveProperties()
        self.slaveProperties.transport_layer = self.transport_name.upper()

    def __enter__(self):
        """Context manager entry part."""
        self.transport.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit part."""
        # if self.connected:
        #    self.disconnect()
        self.close()
        if exc_type is None:
            return
        else:
            self.succeeded = False
            # print("=" * 79)
            # print("Exception while in Context-Manager:\n")
            self.logger.error("".join(traceback.format_exception(exc_type, exc_val, exc_tb)))
            # print("=" * 79)
            # return True

    def _setService(self, service):
        """Records the currently processed service.

        Parameters
        ----------
        service: `pydbc.types.Command`

        Note
        ----
        Internal Function, only to be used by transport-layer.
        """
        self.service = service

    def close(self):
        """Closes transport layer connection."""
        self.transport.policy.finalize()
        self.transport.close()

    # Mandatory Commands.
    @wrapped
    def connect(self, mode=0x00):
        """Build up connection to an XCP slave.

        Before the actual XCP traffic starts a connection is required.

        Parameters
        ----------
        mode : int
            connection mode; default is 0x00 (normal mode)

        Returns
        -------
        :py:obj:`pyxcp.types.ConnectResponse`
            Describes fundamental client properties.

        Note
        ----
        Every XCP slave supports at most one connection,
        more attempts to connect are silently ignored.

        """
        response = self.transport.request(types.Command.CONNECT, mode & 0xFF)

        # First get byte-order
        resultPartial = types.ConnectResponsePartial.parse(response)
        byteOrder = resultPartial.commModeBasic.byteOrder

        result = types.ConnectResponse.parse(response, byteOrder=byteOrder)
        byteOrderPrefix = "<" if byteOrder == types.ByteOrder.INTEL else ">"
        self.slaveProperties.byteOrder = byteOrder
        self.slaveProperties.maxCto = result.maxCto
        self.slaveProperties.maxDto = result.maxDto
        self.slaveProperties.supportsPgm = result.resource.pgm
        self.slaveProperties.supportsStim = result.resource.stim
        self.slaveProperties.supportsDaq = result.resource.daq
        self.slaveProperties.supportsCalpag = result.resource.calpag
        self.slaveProperties.slaveBlockMode = result.commModeBasic.slaveBlockMode
        self.slaveProperties.addressGranularity = result.commModeBasic.addressGranularity
        self.slaveProperties.protocolLayerVersion = result.protocolLayerVersion
        self.slaveProperties.transportLayerVersion = result.transportLayerVersion
        self.slaveProperties.optionalCommMode = result.commModeBasic.optional
        self.slaveProperties.maxWriteDaqMultipleElements = (
            0 if self.slaveProperties.maxCto < 10 else int((self.slaveProperties.maxCto - 2) // 8)
        )
        self.BYTE_pack = makeBytePacker(byteOrderPrefix)
        self.BYTE_unpack = makeByteUnpacker(byteOrderPrefix)
        self.WORD_pack = makeWordPacker(byteOrderPrefix)
        self.WORD_unpack = makeWordUnpacker(byteOrderPrefix)
        self.DWORD_pack = makeDWordPacker(byteOrderPrefix)
        self.DWORD_unpack = makeDWordUnpacker(byteOrderPrefix)
        self.DLONG_pack = makeDLongPacker(byteOrderPrefix)
        self.DLONG_unpack = makeDLongUnpacker(byteOrderPrefix)
        self.slaveProperties.bytesPerElement = None  # Download/Upload commands are using element- not byte-count.
        if self.slaveProperties.addressGranularity == types.AddressGranularity.BYTE:
            self.AG_pack = struct.Struct("<B").pack
            self.AG_unpack = struct.Struct("<B").unpack
            self.slaveProperties.bytesPerElement = 1
        elif self.slaveProperties.addressGranularity == types.AddressGranularity.WORD:
            self.AG_pack = self.WORD_pack
            self.AG_unpack = self.WORD_unpack
            self.slaveProperties.bytesPerElement = 2
        elif self.slaveProperties.addressGranularity == types.AddressGranularity.DWORD:
            self.AG_pack = self.DWORD_pack
            self.AG_unpack = self.DWORD_unpack
            self.slaveProperties.bytesPerElement = 4
            # self.connected = True
        return result

    @wrapped
    def disconnect(self):
        """Releases the connection to the XCP slave.

        Thereafter, no further communication with the slave is possible
        (besides `connect`).


        Note
        -----
        - If DISCONNECT is currently not possible, ERR_CMD_BUSY will be returned.
        - While XCP spec. requires a response, this behavior can be made optional by adding
            - `DISCONNECT_RESPONSE_OPTIONAL = true` (TOML)
            - `"DISCONNECT_RESPONSE_OPTIONAL": true` (JSON)
            to your configuration file.
        """
        if self.disconnect_response_optional:
            response = self.transport.request_optional_response(types.Command.DISCONNECT)
        else:
            response = self.transport.request(types.Command.DISCONNECT)
        # self.connected = False
        return response

    @wrapped
    def getStatus(self):
        """Get current status information of the slave device.

        This includes the status of the resource protection, pending store
        requests and the general status of data acquisition and stimulation.

        Returns
        -------
        :obj:`pyxcp.types.GetStatusResponse`
        """
        response = self.transport.request(types.Command.GET_STATUS)
        result = types.GetStatusResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
        self._setProtectionStatus(result.resourceProtectionStatus)
        return result

    @wrapped
    def synch(self):
        """Synchronize command execution after timeout conditions."""
        response = self.transport.request(types.Command.SYNCH)
        return response

    @wrapped
    def getCommModeInfo(self):
        """Get optional information on different Communication Modes supported
        by the slave.

        Returns
        -------
        :obj:`pyxcp.types.GetCommModeInfoResponse`
        """
        response = self.transport.request(types.Command.GET_COMM_MODE_INFO)
        result = types.GetCommModeInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
        self.slaveProperties.interleavedMode = result.commModeOptional.interleavedMode
        self.slaveProperties.masterBlockMode = result.commModeOptional.masterBlockMode
        self.slaveProperties.maxBs = result.maxBs
        self.slaveProperties.minSt = result.minSt
        self.slaveProperties.queueSize = result.queueSize
        self.slaveProperties.xcpDriverVersionNumber = result.xcpDriverVersionNumber
        return result

    @wrapped
    def getId(self, mode: int):
        """This command is used for automatic session configuration and for
        slave device identification.

        Parameters
        ----------
        mode : int
            The following identification types may be requested:
            - 0        ASCII text
            - 1        ASAM-MC2 filename without path and extension
            - 2        ASAM-MC2 filename with path and extension
            - 3        URL where the ASAM-MC2 file can be found
            - 4        ASAM-MC2 file to upload
            - 128..255 User defined

        Returns
        -------
        :obj:`pydbc.types.GetIDResponse`
        """
        response = self.transport.request(types.Command.GET_ID, mode)
        result = types.GetIDResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
        result.length = self.DWORD_unpack(response[3:7])[0]
        return result

    @wrapped
    def setRequest(self, mode: int, session_configuration_id: int):
        """Request to save to non-volatile memory.

        Parameters
        ----------
        mode : int (bitfield)
            - 1  Request to store calibration data
            - 2  Request to store DAQ list, no resume
            - 4  Request to store DAQ list, resume enabled
            - 8  Request to clear DAQ configuration
        sessionConfigurationId : int

        """
        return self.transport.request(
            types.Command.SET_REQUEST,
            mode,
            session_configuration_id >> 8,
            session_configuration_id & 0xFF,
        )

    @wrapped
    def getSeed(self, first: int, resource: int):
        """Get seed from slave for unlocking a protected resource.

        Parameters
        ----------
        first : int
            - 0 - first part of seed
            - 1 - remaining part
        resource : int
            - Mode = =0 - Resource
            - Mode == 1 - Don't care

        Returns
        -------
        `pydbc.types.GetSeedResponse`
        """
        if self.transport_name == "can":
            # for CAN it might happen that the seed is longer than the max DLC
            # in this case the first byte will be the current remaining seed size
            # followed by the seeds bytes that can fit in the current frame
            # the master must call getSeed several times until the complete seed is received
            response = self.transport.request(types.Command.GET_SEED, first, resource)
            size, seed = response[0], response[1:]
            if size < len(seed):
                seed = seed[:size]
            reply = types.GetSeedResponse.parse(
                types.GetSeedResponse.build({"length": size, "seed": bytes(size)}),
                byteOrder=self.slaveProperties.byteOrder,
            )
            reply.seed = seed
            return reply
        else:
            response = self.transport.request(types.Command.GET_SEED, first, resource)
            return types.GetSeedResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def unlock(self, length: int, key: bytes):
        """Send key to slave for unlocking a protected resource.

        Parameters
        ----------
        length : int
            indicates the (remaining) number of key bytes.
        key : bytes

        Returns
        -------
        :obj:`pydbc.types.ResourceType`

        Note
        ----
        The master has to use :meth:`unlock` in a defined sequence together
        with :meth:`getSeed`. The master only can send an :meth:`unlock` sequence
        if previously there was a :meth:`getSeed` sequence. The master has
        to send the first `unlocking` after a :meth:`getSeed` sequence with
        a Length containing the total length of the key.
        """
        response = self.transport.request(types.Command.UNLOCK, length, *key)
        result = types.ResourceType.parse(response, byteOrder=self.slaveProperties.byteOrder)
        self._setProtectionStatus(result)
        return result

    @wrapped
    def setMta(self, address: int, address_ext: int = 0x00):
        """Set Memory Transfer Address in slave.

        Parameters
        ----------
        address : int
        addressExt : int

        Note
        ----
        The MTA is used by :meth:`buildChecksum`, :meth:`upload`, :meth:`download`, :meth:`downloadNext`,
        :meth:`downloadMax`, :meth:`modifyBits`, :meth:`programClear`, :meth:`program`, :meth:`programNext`
        and :meth:`programMax`.

        """
        self.mta = types.MtaType(address, address_ext)  # Keep track of MTA (needed for error-handling).
        addr = self.DWORD_pack(address)
        return self.transport.request(types.Command.SET_MTA, 0, 0, address_ext, *addr)

    @wrapped
    def upload(self, length: int):
        """Transfer data from slave to master.

        Parameters
        ----------
        length : int
            Number of elements (address granularity).

        Note
        ----
        Adress is set via :meth:`setMta` (Some services like :meth:`getID` also set the MTA).

        Returns
        -------
        bytes
        """
        byte_count = length * self.slaveProperties.bytesPerElement
        response = self.transport.request(types.Command.UPLOAD, length)
        if byte_count > (self.slaveProperties.maxCto - 1):
            block_response = self.transport.block_receive(length_required=(byte_count - len(response)))
            response += block_response
        elif self.transport_name == "can":
            # larger sizes will send in multiple CAN messages
            # each valid message will start with 0xFF followed by the upload bytes
            # the last message might be padded to the required DLC
            rem = byte_count - len(response)
            while rem:
                if len(self.transport.resQueue):
                    data = self.transport.resQueue.popleft()
                    response += data[1 : rem + 1]
                    rem = byte_count - len(response)
                else:
                    short_sleep()
        return response

    @wrapped
    def shortUpload(self, length: int, address: int, address_ext: int = 0x00):
        """Transfer data from slave to master.
        As opposed to :meth:`upload` this service also includes address information.

        Parameters
        ----------
        length : int
            Number of elements (address granularity).
        address : int
        addressExt : int

        Returns
        -------
        bytes
        """
        addr = self.DWORD_pack(address)
        byte_count = length * self.slaveProperties.bytesPerElement
        max_byte_count = self.slaveProperties.maxCto - 1
        if byte_count > max_byte_count:
            self.logger.warn(f"SHORT_UPLOAD: {byte_count} bytes exceeds the maximum value of {max_byte_count}.")
        response = self.transport.request(types.Command.SHORT_UPLOAD, length, 0, address_ext, *addr)
        return response[:byte_count]

    @wrapped
    def buildChecksum(self, blocksize: int):
        """Build checksum over memory range.

        Parameters
        ----------
        blocksize : int

        Returns
        -------
        :obj:`~pyxcp.types.BuildChecksumResponse`

        .. note:: Adress is set via `setMta`

        See Also
        --------
        :mod:`~pyxcp.checksum`
        """
        bs = self.DWORD_pack(blocksize)
        response = self.transport.request(types.Command.BUILD_CHECKSUM, 0, 0, 0, *bs)
        return types.BuildChecksumResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def transportLayerCmd(self, sub_command: int, *data: List[bytes]):
        """Execute transfer-layer specific command.

        Parameters
        ----------
        subCommand : int
        data : bytes

        Note
        ----
        For details refer to XCP specification.
        """
        return self.transport.request_optional_response(types.Command.TRANSPORT_LAYER_CMD, sub_command, *data)

    @wrapped
    def userCmd(self, sub_command: int, data: bytes):
        """Execute proprietary command implemented in your XCP client.

        Parameters
        ----------
        subCommand : int
        data : bytes


        .. note:: For details refer to your XCP client vendor.
        """

        response = self.transport.request(types.Command.USER_CMD, sub_command, *data)
        return response

    @wrapped
    def getVersion(self):
        """Get version information.

        This command returns detailed information about the implemented
        protocol layer version of the XCP slave and the transport layer
        currently in use.

        Returns
        -------
        :obj:`~types.GetVersionResponse`
        """

        response = self.transport.request(types.Command.GET_VERSION)
        result = types.GetVersionResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
        self.slaveProperties.protocolMajor = result.protocolMajor
        self.slaveProperties.protocolMinor = result.protocolMinor
        self.slaveProperties.transportMajor = result.transportMajor
        self.slaveProperties.transportMinor = result.transportMinor
        return result

    def fetch(self, length: int, limit_payload: int = None):  # TODO: pull
        """Convenience function for data-transfer from slave to master
        (Not part of the XCP Specification).

        Parameters
        ----------
        length : int
        limitPayload : int
            transfer less bytes then supported by transport-layer

        Returns
        -------
        bytes

        Note
        ----
        address is not included because of services implicitly setting address information like :meth:`getID` .
        """
        if limit_payload and limit_payload < 8:
            raise ValueError(f"Payload must be at least 8 bytes - given: {limit_payload}")

        slave_block_mode = self.slaveProperties.slaveBlockMode
        if slave_block_mode:
            max_payload = 255
        else:
            max_payload = self.slaveProperties.maxCto - 1
        payload = min(limit_payload, max_payload) if limit_payload else max_payload
        chunk_size = payload
        chunks = range(length // chunk_size)
        remaining = length % chunk_size
        result = []
        for _ in chunks:
            data = self.upload(chunk_size)
            result.extend(data[:chunk_size])
        if remaining:
            data = self.upload(remaining)
            result.extend(data[:remaining])
        return bytes(result)

    pull = fetch  # fetch() may be completely replaced by pull() someday.

    def push(self, address: int, address_ext: int, data: bytes, callback: Optional[Callable] = None):
        """Convenience function for data-transfer from master to slave.
        (Not part of the XCP Specification).

        Parameters
        ----------
        address: int

        data : bytes
            Arbitrary number of bytes.

        Returns
        -------
        """
        self._generalized_downloader(
            address=address,
            address_ext=address_ext,
            data=data,
            maxCto=self.slaveProperties.maxCto,
            maxBs=self.slaveProperties.maxBs,
            minSt=self.slaveProperties.minSt,
            master_block_mode=self.slaveProperties.masterBlockMode,
            dl_func=self.download,
            dl_next_func=self.downloadNext,
            callback=callback,
        )

    def flash_program(self, address: int, data: bytes, callback: Optional[Callable] = None):
        """Convenience function for flash programing.
        (Not part of the XCP Specification).

        Parameters
        ----------
        address: int

        data : bytes
            Arbitrary number of bytes.

        Returns
        -------
        """
        self._generalized_downloader(
            address=address,
            data=data,
            maxCto=self.slaveProperties.pgmProcessor.maxCtoPgm,
            maxBs=self.slaveProperties.pgmProcessor.maxBsPgm,
            minSt=self.slaveProperties.pgmProcessor.minStPgm,
            master_block_mode=self.slaveProperties.pgmProcessor.masterBlockMode,
            dl_func=self.program,
            dl_next_func=self.programNext,
            callback=callback,
        )

    def _generalized_downloader(
        self,
        address: int,
        address_ext: int,
        data: bytes,
        maxCto: int,
        maxBs: int,
        minSt: int,
        master_block_mode: bool,
        dl_func,
        dl_next_func,
        callback=None,
    ):
        """ """
        self.setMta(address, address_ext)
        minSt /= 10000.0
        block_downloader = functools.partial(
            self._block_downloader,
            dl_func=dl_func,
            dl_next_func=dl_next_func,
            minSt=minSt,
        )
        total_length = len(data)
        if master_block_mode:
            max_payload = min(maxBs * (maxCto - 2), 255)
        else:
            max_payload = maxCto - 2
        offset = 0
        if master_block_mode:
            remaining = total_length
            blocks = range(total_length // max_payload)
            percent_complete = 1
            remaining_block_size = total_length % max_payload
            for _ in blocks:
                block = data[offset : offset + max_payload]
                block_downloader(block)
                offset += max_payload
                remaining -= max_payload
                if callback and remaining <= total_length - (total_length / 100) * percent_complete:
                    callback(percent_complete)
                    percent_complete += 1
            if remaining_block_size:
                block = data[offset : offset + remaining_block_size]
                block_downloader(block)
                if callback:
                    callback(percent_complete)
        else:
            chunk_size = max_payload
            chunks = range(total_length // chunk_size)
            remaining = total_length % chunk_size
            percent_complete = 1
            callback_remaining = total_length
            for _ in chunks:
                block = data[offset : offset + max_payload]
                dl_func(block, max_payload, last=False)
                offset += max_payload
                callback_remaining -= chunk_size
                if callback and callback_remaining <= total_length - (total_length / 100) * percent_complete:
                    callback(percent_complete)
                    percent_complete += 1
            if remaining:
                block = data[offset : offset + remaining]
                dl_func(block, remaining, last=True)
                if callback:
                    callback(percent_complete)

    def _block_downloader(
        self, data: bytes, dl_func: Optional[Callable] = None, dl_next_func: Optional[Callable] = None, minSt: int = 0
    ):
        """Re-usable block downloader.

        Parameters
        ----------
        data : bytes
            Arbitrary number of bytes.

        dl_func: method
            usually :meth: `download` or :meth:`program`

        dl_next_func: method
            usually :meth: `downloadNext` or :meth:`programNext`

        minSt: int
            Minimum separation time of frames.
        """
        length = len(data)
        max_packet_size = self.slaveProperties.maxCto - 2  # Command ID + Length
        packets = range(length // max_packet_size)
        offset = 0
        remaining = length % max_packet_size
        remaining_block_size = length
        index = 0
        for index in packets:
            packet_data = data[offset : offset + max_packet_size]
            last = (remaining_block_size - max_packet_size) == 0
            if index == 0:
                dl_func(packet_data, length, last)  # Transmit the complete length in the first CTO.
            else:
                dl_next_func(packet_data, remaining_block_size, last)
            offset += max_packet_size
            remaining_block_size -= max_packet_size
            delay(minSt)
        if remaining:
            packet_data = data[offset : offset + remaining]
            if index == 0:
                # length of data is smaller than maxCto - 2
                dl_func(packet_data, remaining, last=True)
            else:
                dl_next_func(packet_data, remaining, last=True)
            delay(minSt)

    @wrapped
    def download(self, data: bytes, block_mode_length: Optional[int] = None, last: bool = False):
        """Transfer data from master to slave.

        Parameters
        ----------
        data : bytes
            Data to send to slave.
        block_mode_length : int or None
            for block mode, the download request must contain the length of the whole block,
            not just the length in the current packet. The whole block length can be given here for block-mode
            transfers. For normal mode, the length indicates the actual packet's payload length.

        Note
        ----
        Adress is set via :meth:`setMta`
        """

        if block_mode_length is None or last:
            # standard mode
            length = len(data)
            response = self.transport.request(types.Command.DOWNLOAD, length, *data)
            return response
        else:
            # block mode
            if not isinstance(block_mode_length, int):
                raise TypeError("block_mode_length must be int!")
            self.transport.block_request(types.Command.DOWNLOAD, block_mode_length, *data)
            return None

    @wrapped
    def downloadNext(self, data: bytes, remaining_block_length: int, last: bool = False):
        """Transfer data from master to slave (block mode).

        Parameters
        ----------
        data : bytes
        remainingBlockLength : int
            This parameter has to be given the remaining length in the block
        last : bool
            The block mode implementation shall indicate the last packet in the block with this parameter, because
            the slave device will send the response after this.
        """

        if last:
            # last DOWNLOAD_NEXT packet in a block: the slave device has to send the response after this.
            response = self.transport.request(types.Command.DOWNLOAD_NEXT, remaining_block_length, *data)
            return response
        else:
            # the slave device won't respond to consecutive DOWNLOAD_NEXT packets in block mode,
            # so we must not wait for any response
            self.transport.block_request(types.Command.DOWNLOAD_NEXT, remaining_block_length, *data)
            return None

    @wrapped
    def downloadMax(self, data: bytes):
        """Transfer data from master to slave (fixed size).

        Parameters
        ----------
        data : bytes
        """
        return self.transport.request(types.Command.DOWNLOAD_MAX, *data)

    @wrapped
    def shortDownload(self, address: int, address_ext: int, data: bytes):
        length = len(data)
        addr = self.DWORD_pack(address)
        return self.transport.request(types.Command.SHORT_DOWNLOAD, length, 0, address_ext, *addr, *data)

    @wrapped
    def modifyBits(self, shift_value: int, and_mask: int, xor_mask: int):
        # A = ( (A) & ((~((dword)(((word)~MA)<<S))) )^((dword)(MX<<S)) )
        am = self.WORD_pack(and_mask)
        xm = self.WORD_pack(xor_mask)
        return self.transport.request(types.Command.MODIFY_BITS, shift_value, *am, *xm)

    # Page Switching Commands (PAG)
    @wrapped
    def setCalPage(self, mode: int, logical_data_segment: int, logical_data_page: int):
        """Set calibration page.

        Parameters
        ----------
        mode : int (bitfield)
            - 0x01 - The given page will be used by the slave device application.
            - 0x02 - The slave device XCP driver will access the given page.
            - 0x80 - The logical segment number is ignored. The command applies to all segments
        logicalDataSegment : int
        logicalDataPage : int
        """
        return self.transport.request(types.Command.SET_CAL_PAGE, mode, logical_data_segment, logical_data_page)

    @wrapped
    def getCalPage(self, mode: int, logical_data_segment: int):
        """Get calibration page

        Parameters
        ----------
        mode : int
        logicalDataSegment : int
        """
        response = self.transport.request(types.Command.GET_CAL_PAGE, mode, logical_data_segment)
        return response[2]

    @wrapped
    def getPagProcessorInfo(self):
        """Get general information on PAG processor.

        Returns
        -------
        `pydbc.types.GetPagProcessorInfoResponse`
        """
        response = self.transport.request(types.Command.GET_PAG_PROCESSOR_INFO)
        return types.GetPagProcessorInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def getSegmentInfo(self, mode: int, segment_number: int, segment_info: int, mapping_index: int):
        """Get specific information for a segment.

        Parameters
        ----------
        mode : int
            - 0 = get basic address info for this segment
            - 1 = get standard info for this segment
            - 2 = get address mapping info for this segment

        segmentNumber : int
        segmentInfo : int
            Mode 0:
                - 0 = address
                - 1 = length

            Mode 1:
                - don't care

            Mode 2:
                - 0 = source address
                - 1 = destination address
                - 2 = length address

        mappingIndex : int
            - Mode 0: don't care
            - Mode 1: don't care
            - Mode 2: identifier for address mapping range that mapping_info belongs to.

        """
        response = self.transport.request(
            types.Command.GET_SEGMENT_INFO,
            mode,
            segment_number,
            segment_info,
            mapping_index,
        )
        if mode == 0:
            return types.GetSegmentInfoMode0Response.parse(response, byteOrder=self.slaveProperties.byteOrder)
        elif mode == 1:
            return types.GetSegmentInfoMode1Response.parse(response, byteOrder=self.slaveProperties.byteOrder)
        elif mode == 2:
            return types.GetSegmentInfoMode2Response.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def getPageInfo(self, segment_number: int, page_number: int):
        """Get specific information for a page.

        Parameters
        ----------
        segmentNumber : int
        pageNumber : int
        """
        response = self.transport.request(types.Command.GET_PAGE_INFO, 0, segment_number, page_number)
        return (
            types.PageProperties.parse(bytes([response[0]]), byteOrder=self.slaveProperties.byteOrder),
            response[1],
        )

    @wrapped
    def setSegmentMode(self, mode: int, segment_number: int):
        """Set mode for a segment.

        Parameters
        ----------
        mode : int (bitfield)
            1 = enable FREEZE Mode
        segmentNumber : int
        """
        return self.transport.request(types.Command.SET_SEGMENT_MODE, mode, segment_number)

    @wrapped
    def getSegmentMode(self, segment_number: int):
        """Get mode for a segment.

        Parameters
        ----------
        segmentNumber : int
        """
        response = self.transport.request(types.Command.GET_SEGMENT_MODE, 0, segment_number)
        if response:
            return response[1]

    @wrapped
    def copyCalPage(self, src_segment: int, src_page: int, dst_segment: int, dst_page: int):
        """Copy page.

        Parameters
        ----------
        srcSegment : int
        srcPage : int
        dstSegment : int
        dstPage : int
        """
        return self.transport.request(types.Command.COPY_CAL_PAGE, src_segment, src_page, dst_segment, dst_page)

    # DAQ

    @wrapped
    def setDaqPtr(self, daq_list_number: int, odt_number: int, odt_entry_number: int):
        self.currentDaqPtr = types.DaqPtr(daq_list_number, odt_number, odt_entry_number)  # Needed for errorhandling.
        daq_list = self.WORD_pack(daq_list_number)
        response = self.transport.request(types.Command.SET_DAQ_PTR, 0, *daq_list, odt_number, odt_entry_number)
        self.stim.setDaqPtr(daq_list_number, odt_number, odt_entry_number)
        return response

    @wrapped
    def clearDaqList(self, daq_list_number: int):
        """Clear DAQ list configuration.

        Parameters
        ----------
        daqListNumber : int
        """
        daq_list = self.WORD_pack(daq_list_number)
        result = self.transport.request(types.Command.CLEAR_DAQ_LIST, 0, *daq_list)
        self.stim.clearDaqList(daq_list_number)
        return result

    @wrapped
    def writeDaq(self, bit_offset: int, entry_size: int, address_ext: int, address: int):
        """Write element in ODT entry.

        Parameters
        ----------
        bitOffset : int
            Position of bit in 32-bit variable referenced by the address and
            extension below
        entrySize : int
        addressExt : int
        address : int
        """
        addr = self.DWORD_pack(address)
        result = self.transport.request(types.Command.WRITE_DAQ, bit_offset, entry_size, address_ext, *addr)
        self.stim.writeDaq(bit_offset, entry_size, address_ext, address)
        return result

    @wrapped
    def setDaqListMode(self, mode: int, daq_list_number: int, event_channel_number: int, prescaler: int, priority: int):
        dln = self.WORD_pack(daq_list_number)
        ecn = self.WORD_pack(event_channel_number)
        self.stim.setDaqListMode(mode, daq_list_number, event_channel_number, prescaler, priority)
        return self.transport.request(types.Command.SET_DAQ_LIST_MODE, mode, *dln, *ecn, prescaler, priority)

    @wrapped
    def getDaqListMode(self, daq_list_number: int):
        """Get mode from DAQ list.

        Parameters
        ----------
        daqListNumber : int

        Returns
        -------
        `pyxcp.types.GetDaqListModeResponse`
        """
        dln = self.WORD_pack(daq_list_number)
        response = self.transport.request(types.Command.GET_DAQ_LIST_MODE, 0, *dln)
        return types.GetDaqListModeResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def startStopDaqList(self, mode: int, daq_list_number: int):
        """Start /stop/select DAQ list.

        Parameters
        ----------
        mode : int
            0 = stop
            1 = start
            2 = select
        daqListNumber : int
        """
        dln = self.WORD_pack(daq_list_number)
        response = self.transport.request(types.Command.START_STOP_DAQ_LIST, mode, *dln)
        self.stim.startStopDaqList(mode, daq_list_number)
        first_pid = types.StartStopDaqListResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
        self.stim.set_first_pid(daq_list_number, first_pid.firstPid)
        return first_pid

    @wrapped
    def startStopSynch(self, mode: int):
        """Start/stop DAQ lists (synchronously).

        Parameters
        ----------
        mode : int
            0 = stop all
            1 = start selected
            2 = stop selected
        """
        res = self.transport.request(types.Command.START_STOP_SYNCH, mode)
        self.stim.startStopSynch(mode)
        return res

    @wrapped
    def writeDaqMultiple(self, daq_elements: dict):
        """Write multiple elements in ODT.

        Parameters
        ----------
        daqElements : list of `dict` containing the following keys: *bitOffset*, *size*, *address*, *addressExt*.
        """
        if len(daq_elements) > self.slaveProperties.maxWriteDaqMultipleElements:
            raise ValueError(f"At most {self.slaveProperties.maxWriteDaqMultipleElements} daqElements are permitted.")
        data = bytearray()
        data.append(len(daq_elements))

        for daq_element in daq_elements:
            data.extend(types.DaqElement.build(daq_element, byteOrder=self.slaveProperties.byteOrder))

        return self.transport.request(types.Command.WRITE_DAQ_MULTIPLE, *data)

    # optional
    @wrapped
    def getDaqClock(self):
        """Get DAQ clock from slave.

        Returns
        -------
        int
            Current timestamp, format specified by `getDaqResolutionInfo`
        """
        response = self.transport.request(types.Command.GET_DAQ_CLOCK)
        result = types.GetDaqClockResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
        return result.timestamp

    @wrapped
    def readDaq(self):
        """Read element from ODT entry.

        Returns
        -------
        `pyxcp.types.ReadDaqResponse`
        """
        response = self.transport.request(types.Command.READ_DAQ)
        return types.ReadDaqResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def getDaqProcessorInfo(self):
        """Get general information on DAQ processor.

        Returns
        -------
        `pyxcp.types.GetDaqProcessorInfoResponse`
        """
        response = self.transport.request(types.Command.GET_DAQ_PROCESSOR_INFO)
        return types.GetDaqProcessorInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def getDaqResolutionInfo(self):
        """Get general information on DAQ processing resolution.

        Returns
        -------
        `pyxcp.types.GetDaqResolutionInfoResponse`
        """
        response = self.transport.request(types.Command.GET_DAQ_RESOLUTION_INFO)
        return types.GetDaqResolutionInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def getDaqListInfo(self, daq_list_number: int):
        """Get specific information for a DAQ list.

        Parameters
        ----------
        daqListNumber : int
        """
        dln = self.WORD_pack(daq_list_number)
        response = self.transport.request(types.Command.GET_DAQ_LIST_INFO, 0, *dln)
        return types.GetDaqListInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def getDaqEventInfo(self, event_channel_number: int):
        """Get specific information for an event channel.

        Parameters
        ----------
        eventChannelNumber : int

        Returns
        -------
        `pyxcp.types.GetEventChannelInfoResponse`
        """
        ecn = self.WORD_pack(event_channel_number)
        response = self.transport.request(types.Command.GET_DAQ_EVENT_INFO, 0, *ecn)
        return types.GetEventChannelInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def dtoCtrProperties(self, modifier: int, event_channel: int, related_event_channel: int, mode: int):
        """DTO CTR properties

        Parameters
        ----------
        modifier :
        eventChannel : int
        relatedEventChannel : int
        mode :

        Returns
        -------
        `pyxcp.types.DtoCtrPropertiesResponse`
        """
        data = bytearray()
        data.append(modifier)
        data.extend(self.WORD_pack(event_channel))
        data.extend(self.WORD_pack(related_event_channel))
        data.append(mode)
        response = self.transport.request(types.Command.DTO_CTR_PROPERTIES, *data)
        return types.DtoCtrPropertiesResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def setDaqPackedMode(
        self, daq_list_number: int, daq_packed_mode: int, dpm_timestamp_mode: int = None, dpm_sample_count: int = None
    ):
        """Set DAQ List Packed Mode.

        Parameters
        ----------
        daqListNumber : int
        daqPackedMode : int
        """
        params = []
        dln = self.WORD_pack(daq_list_number)
        params.extend(dln)
        params.append(daq_packed_mode)

        if daq_packed_mode == 1 or daq_packed_mode == 2:
            params.append(dpm_timestamp_mode)
            dsc = self.WORD_pack(dpm_sample_count)
            params.extend(dsc)

        return self.transport.request(types.Command.SET_DAQ_PACKED_MODE, *params)

    @wrapped
    def getDaqPackedMode(self, daq_list_number: int):
        """Get DAQ List Packed Mode.

        This command returns information of the currently active packed mode of
        the addressed DAQ list.

        Parameters
        ----------
        daqListNumber : int
        """
        dln = self.WORD_pack(daq_list_number)
        response = self.transport.request(types.Command.GET_DAQ_PACKED_MODE, *dln)
        return types.GetDaqPackedModeResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    # dynamic
    @wrapped
    def freeDaq(self):
        """Clear dynamic DAQ configuration."""
        result = self.transport.request(types.Command.FREE_DAQ)
        self.stim.freeDaq()
        return result

    @wrapped
    def allocDaq(self, daq_count: int):
        """Allocate DAQ lists.

        Parameters
        ----------
        daqCount : int
            number of DAQ lists to be allocated
        """
        dq = self.WORD_pack(daq_count)
        result = self.transport.request(types.Command.ALLOC_DAQ, 0, *dq)
        self.stim.allocDaq(daq_count)
        return result

    @wrapped
    def allocOdt(self, daq_list_number: int, odt_count: int):
        dln = self.WORD_pack(daq_list_number)
        result = self.transport.request(types.Command.ALLOC_ODT, 0, *dln, odt_count)
        self.stim.allocOdt(daq_list_number, odt_count)
        return result

    @wrapped
    def allocOdtEntry(self, daq_list_number: int, odt_number: int, odt_entries_count: int):
        dln = self.WORD_pack(daq_list_number)
        result = self.transport.request(types.Command.ALLOC_ODT_ENTRY, 0, *dln, odt_number, odt_entries_count)
        self.stim.allocOdtEntry(daq_list_number, odt_number, odt_entries_count)
        return result

    # PGM
    @wrapped
    def programStart(self):
        """Indicate the beginning of a programming sequence.

        Returns
        -------
        `pyxcp.types.ProgramStartResponse`
        """
        response = self.transport.request(types.Command.PROGRAM_START)
        result = types.ProgramStartResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
        self.slaveProperties.pgmProcessor.commModePgm = result.commModePgm
        self.slaveProperties.pgmProcessor.maxCtoPgm = result.maxCtoPgm
        self.slaveProperties.pgmProcessor.maxBsPgm = result.maxBsPgm
        self.slaveProperties.pgmProcessor.minStPgm = result.minStPgm
        self.slaveProperties.pgmProcessor.queueSizePgm = result.queueSizePgm
        self.slaveProperties.pgmProcessor.slaveBlockMode = result.commModePgm.slaveBlockMode
        self.slaveProperties.pgmProcessor.interleavedMode = result.commModePgm.interleavedMode
        self.slaveProperties.pgmProcessor.masterBlockMode = result.commModePgm.masterBlockMode
        return result

    @wrapped
    def programClear(self, mode: int, clear_range: int):
        """Clear a part of non-volatile memory.

        Parameters
        ----------
        mode : int
            0x00 = the absolute access mode is active (default)
            0x01 = the functional access mode is active
        clearRange : int
        """
        cr = self.DWORD_pack(clear_range)
        response = self.transport.request(types.Command.PROGRAM_CLEAR, mode, 0, 0, *cr)
        # ERR_ACCESS_LOCKED
        return response

    @wrapped
    def program(self, data: bytes, block_length: int, last: bool = False):
        """Parameters
        ----------
        data : bytes
            Data to send to slave.
        block_mode_length : int
            the program request must contain the length of the whole block, not just the length
            in the current packet.
        last : bool
            Indicates that this is the only packet in the block, because
            the slave device will send the response after this.

        Note
        ----
        Adress is set via :meth:`setMta`
        """
        # d = bytearray()
        # d.append(len(data))
        # if self.slaveProperties.addressGranularity == types.AddressGranularity.DWORD:
        #    d.extend(b"\x00\x00")  # alignment bytes
        # for e in data:
        #    d.extend(self.AG_pack(e))
        if last:
            # last PROGRAM_NEXT packet in a block: the slave device has to send the response after this.
            response = self.transport.request(types.Command.PROGRAM, block_length, *data)
            return response
        else:
            # the slave device won't respond to consecutive PROGRAM_NEXT packets in block mode,
            # so we must not wait for any response
            self.transport.block_request(types.Command.PROGRAM, block_length, *data)
            return None

    @wrapped
    def programReset(self, wait_for_optional_response=True):
        """Indicate the end of a programming sequence."""
        if wait_for_optional_response:
            return self.transport.request_optional_response(types.Command.PROGRAM_RESET)
        else:
            return self.transport.request(types.Command.PROGRAM_RESET)

    @wrapped
    def getPgmProcessorInfo(self):
        """Get general information on PGM processor."""
        response = self.transport.request(types.Command.GET_PGM_PROCESSOR_INFO)
        result = types.GetPgmProcessorInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)
        self.slaveProperties.pgmProcessor.pgmProperties = result.pgmProperties
        self.slaveProperties.pgmProcessor.maxSector = result.maxSector
        return result

    @wrapped
    def getSectorInfo(self, mode: int, sector_number: int):
        """Get specific information for a sector."""
        response = self.transport.request(types.Command.GET_SECTOR_INFO, mode, sector_number)
        if mode == 0 or mode == 1:
            return types.GetSectorInfoResponseMode01.parse(response, byteOrder=self.slaveProperties.byteOrder)
        elif mode == 2:
            return types.GetSectorInfoResponseMode2.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def programPrepare(self, codesize):
        """Prepare non-volatile memory programming."""
        cs = self.WORD_pack(codesize)
        return self.transport.request(types.Command.PROGRAM_PREPARE, 0x00, *cs)

    @wrapped
    def programFormat(self, compression_method: int, encryption_method: int, programming_method: int, access_method: int):
        return self.transport.request(
            types.Command.PROGRAM_FORMAT,
            compression_method,
            encryption_method,
            programming_method,
            access_method,
        )

    @wrapped
    def programNext(self, data: bytes, remaining_block_length: int, last: bool = False):
        # d = bytearray()
        # d.append(len(data))
        # if self.slaveProperties.addressGranularity == types.AddressGranularity.DWORD:
        #    d.extend(b"\x00\x00")  # alignment bytes
        # for e in data:
        #    d.extend(self.AG_pack(e))
        if last:
            # last PROGRAM_NEXT packet in a block: the slave device has to send the response after this.
            response = self.transport.request(types.Command.PROGRAM_NEXT, remaining_block_length, *data)
            return response
        else:
            # the slave device won't respond to consecutive PROGRAM_NEXT packets in block mode,
            # so we must not wait for any response
            self.transport.block_request(types.Command.PROGRAM_NEXT, remaining_block_length, *data)
            return None

    @wrapped
    def programMax(self, data: bytes):
        d = bytearray()
        if self.slaveProperties.addressGranularity == types.AddressGranularity.WORD:
            d.extend(b"\x00")  # alignment bytes
        elif self.slaveProperties.addressGranularity == types.AddressGranularity.DWORD:
            d.extend(b"\x00\x00\x00")  # alignment bytes
        for e in data:
            d.extend(self.AG_pack(e))
        return self.transport.request(types.Command.PROGRAM_MAX, *d)

    @wrapped
    def programVerify(self, ver_mode: int, ver_type: int, ver_value: int):
        data = bytearray()
        data.extend(self.WORD_pack(ver_type))
        data.extend(self.DWORD_pack(ver_value))
        return self.transport.request(types.Command.PROGRAM_VERIFY, ver_mode, *data)

    # DBG

    @wrapped
    def dbgAttach(self):
        """Returns detailed information about the implemented version of the SW-DBG feature of the XCP slave

        Returns
        -------
        `pyxcp.types.DbgAttachResponse`
        """
        response = self.transport.request(types.Command.DBG_ATTACH)
        return types.DbgAttachResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def dbgGetVendorInfo(self):
        """"""
        response = self.transport.request(types.Command.DBG_GET_VENDOR_INFO)
        return types.DbgGetVendorInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def dbgGetModeInfo(self):
        """"""
        response = self.transport.request(types.Command.DBG_GET_MODE_INFO)
        return types.DbgGetModeInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def dbgGetJtagId(self):
        """"""
        response = self.transport.request(types.Command.DBG_GET_JTAG_ID)
        return types.DbgGetJtagIdResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def dbgHaltAfterReset(self):
        """"""
        return self.transport.request(types.Command.DBG_HALT_AFTER_RESET)

    @wrapped
    def dbgGetHwioInfo(self, index: int):
        """"""
        response = self.transport.request(types.Command.DBG_GET_HWIO_INFO, index)
        return types.DbgGetHwioInfoResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def dbgSetHwioEvent(self, index: int, trigger: int):
        """"""
        return self.transport.request(types.Command.DBG_SET_HWIO_EVENT, index, trigger)

    @wrapped
    def dbgHwioControl(self, pins):
        """"""
        d = bytearray()
        d.extend(self.BYTE_pack(len(pins)))
        for p in pins:
            d.extend(self.BYTE_pack(p[0]))  # index
            d.extend(self.BYTE_pack(p[1]))  # state
            d.extend(self.WORD_pack(p[2]))  # frequency

        response = self.transport.request(types.Command.DBG_HWIO_CONTROL, *d)
        return types.DbgHwioControlResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def dbgExclusiveTargetAccess(self, mode: int, context: int):
        """"""
        return self.transport.request(types.Command.DBG_EXCLUSIVE_TARGET_ACCESS, mode, context)

    @wrapped
    def dbgSequenceMultiple(self, mode: int, num: int, *seq):
        """"""
        response = self.transport.request(types.Command.DBG_SEQUENCE_MULTIPLE, mode, self.WORD_pack(num), *seq)
        return types.DbgSequenceMultipleResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def dbgLlt(self, num: int, mode: int, *llts):
        """"""
        response = self.transport.request(types.Command.DBG_LLT, num, mode, *llts)
        return types.DbgLltResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def dbgReadModifyWrite(self, tri: int, width: int, address: int, mask: int, data: int):
        """"""
        d = bytearray()
        d.extend(b"\x00")
        d.append(tri)
        d.append(width)
        d.extend(b"\x00\x00")
        d.extend(self.DLONG_pack(address))
        if width == 0x01:
            d.extend(self.BYTE_pack(mask))
            d.extend(self.BYTE_pack(data))
        elif width == 0x02:
            d.extend(self.WORD_pack(mask))
            d.extend(self.WORD_pack(data))
        elif width == 0x04:
            d.extend(self.DWORD_pack(mask))
            d.extend(self.DWORD_pack(data))
        elif width == 0x08:
            d.extend(self.DLONG_pack(mask))
            d.extend(self.DLONG_pack(data))
        response = self.transport.request(types.Command.DBG_READ_MODIFY_WRITE, *d)
        return types.DbgReadModifyWriteResponse.parse(response, byteOrder=self.slaveProperties.byteOrder, width=width)

    @wrapped
    def dbgWrite(self, tri: int, width: int, address: int, data):
        """"""
        d = bytearray()
        d.extend(b"\x00")
        d.append(tri)
        self._dbg_width = width
        d.append(width)
        d.extend(self.WORD_pack(len(data)))
        d.extend(self.DLONG_pack(address))
        for da in data:
            if width == 0x01:
                d.extend(self.BYTE_pack(da))
            elif width == 0x02:
                d.extend(self.WORD_pack(da))
            elif width == 0x04:
                d.extend(self.DWORD_pack(da))
            elif width == 0x08:
                d.extend(self.DLONG_pack(da))
        return self.transport.request(types.Command.DBG_WRITE, *d)

    @wrapped
    def dbgWriteNext(self, num: int, data: int):
        """"""
        d = bytearray()
        d.extend(b"\x00")
        d.extend(self.WORD_pack(num))
        d.extend(b"\x00\x00")
        for i in range(num):
            if self._dbg_width == 0x01:
                d.extend(self.BYTE_pack(data[i]))
            elif self._dbg_width == 0x02:
                d.extend(self.WORD_pack(data[i]))
            elif self._dbg_width == 0x04:
                d.extend(self.DWORD_pack(data[i]))
            elif self._dbg_width == 0x08:
                d.extend(self.DLONG_pack(data[i]))
        return self.transport.request(types.Command.DBG_WRITE_NEXT, *d)

    @wrapped
    def dbgWriteCan1(self, tri: int, address: int):
        """"""
        d = bytearray()
        d.extend(self.BYTE_pack(tri))
        d.extend(self.DWORD_pack(address))
        return self.transport.request(types.Command.DBG_WRITE_CAN1, *d)

    @wrapped
    def dbgWriteCan2(self, width: int, num: int):
        """"""
        d = bytearray()
        self._dbg_width = width
        d.append(width)
        d.extend(self.BYTE_pack(num))
        return self.transport.request(types.Command.DBG_WRITE_CAN2, *d)

    @wrapped
    def dbgWriteCanNext(self, num: int, data: int):
        """"""
        d = bytearray()
        d.extend(self.BYTE_pack(num))
        for i in range(num):
            if self._dbg_width == 0x01:
                d.extend(self.BYTE_pack(data[i]))
            elif self._dbg_width == 0x02:
                d.extend(self.WORD_pack(data[i]))
            elif self._dbg_width == 0x04:
                d.extend(self.DWORD_pack(data[i]))
            elif self._dbg_width == 0x08:
                d.extend(self.DLONG_pack(data[i]))
        return self.transport.request(types.Command.DBG_WRITE_CAN_NEXT, *d)

    @wrapped
    def dbgRead(self, tri: int, width: int, num: int, address: int):
        """"""
        d = bytearray()
        d.extend(b"\x00")
        d.extend(self.BYTE_pack(tri))
        self._dbg_width = width
        d.extend(self.BYTE_pack(width))
        d.extend(self.WORD_pack(num))
        d.extend(self.DLONG_pack(address))
        response = self.transport.request(types.Command.DBG_READ, *d)
        return types.DbgReadResponse.parse(response, byteOrder=self.slaveProperties.byteOrder, width=width)

    @wrapped
    def dbgReadCan1(self, tri: int, address: int):
        """"""
        d = bytearray()
        d.extend(self.BYTE_pack(tri))
        d.extend(self.DWORD_pack(address))
        return self.transport.request(types.Command.DBG_READ_CAN1, *d)

    @wrapped
    def dbgReadCan2(self, width: int, num: int):
        """"""
        d = bytearray()
        self._dbg_width = width
        d.extend(self.BYTE_pack(width))
        d.extend(self.BYTE_pack(num))
        return self.transport.request(types.Command.DBG_READ_CAN2, *d)

    @wrapped
    def dbgGetTriDescTbl(self):
        """"""
        response = self.transport.request(types.Command.DBG_GET_TRI_DESC_TBL, b"\x00\x00\x00\x00\x00")
        return types.DbgGetTriDescTblResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def dbgLlbt(self, data):
        """"""
        d = bytearray()
        d.extend(b"\x00")
        d.extend(self.WORD_pack(len(data)))
        for b in data:
            d.extend(self.BYTE_pack(b))
        response = self.transport.request(types.Command.DBG_LLBT, d)
        return types.DbgLlbtResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    @wrapped
    def timeCorrelationProperties(self, set_properties: int, get_properties_request: int, cluster_id: int):
        response = self.transport.request(
            types.Command.TIME_CORRELATION_PROPERTIES, set_properties, get_properties_request, 0, *self.WORD_pack(cluster_id)
        )
        return types.TimeCorrelationPropertiesResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    # Transport layer commands / CAN.

    @broadcasted
    @wrapped
    def getSlaveID(self, mode: int):
        response = self.transportLayerCmd(types.TransportLayerCommands.GET_SLAVE_ID, ord("X"), ord("C"), ord("P"), mode)
        return types.GetSlaveIdResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    def getDaqId(self, daq_list_number: int):
        response = self.transportLayerCmd(types.TransportLayerCommands.GET_DAQ_ID, *self.WORD_pack(daq_list_number))
        # if response:
        return types.GetDaqIdResponse.parse(response, byteOrder=self.slaveProperties.byteOrder)

    def setDaqId(self, daq_list_number: int, identifier: int):
        response = self.transportLayerCmd(
            types.TransportLayerCommands.SET_DAQ_ID, *self.WORD_pack(daq_list_number), *self.DWORD_pack(identifier)
        )
        return response

    # Convenience Functions.
    def verify(self, addr, length):
        """Convenience function for verification of a data-transfer from slave
        to master (Not part of the XCP Specification).

        Parameters
        ----------
        addr : int
        length : int

        Returns
        -------
        bool
        """
        self.setMta(addr)
        cs = self.buildChecksum(length)
        self.logger.debug(f"BuildChecksum return'd: 0x{cs.checksum:08X} [{cs.checksumType}]")
        self.setMta(addr)
        data = self.fetch(length)
        cc = checksum.check(data, cs.checksumType)
        self.logger.debug(f"Our checksum          : 0x{cc:08X}")
        return cs.checksum == cc

    def getDaqInfo(self, include_event_lists=True):
        """Get DAQ information: processor, resolution, events."""
        result = {}
        dpi = self.getDaqProcessorInfo()
        processorInfo = {
            "minDaq": dpi["minDaq"],
            "maxDaq": dpi["maxDaq"],
            "properties": {
                "configType": dpi["daqProperties"]["daqConfigType"],
                "overloadEvent": dpi["daqProperties"]["overloadEvent"],
                "overloadMsb": dpi["daqProperties"]["overloadMsb"],
                "prescalerSupported": dpi["daqProperties"]["prescalerSupported"],
                "pidOffSupported": dpi["daqProperties"]["pidOffSupported"],
                "timestampSupported": dpi["daqProperties"]["timestampSupported"],
                "bitStimSupported": dpi["daqProperties"]["bitStimSupported"],
                "resumeSupported": dpi["daqProperties"]["resumeSupported"],
            },
            "keyByte": {
                "identificationField": dpi["daqKeyByte"]["Identification_Field"],
                "addressExtension": dpi["daqKeyByte"]["Address_Extension"],
                "optimisationType": dpi["daqKeyByte"]["Optimisation_Type"],
            },
        }
        result["processor"] = processorInfo

        dri = self.getDaqResolutionInfo()
        resolutionInfo = {
            "timestampTicks": dri["timestampTicks"],
            "maxOdtEntrySizeDaq": dri["maxOdtEntrySizeDaq"],
            "maxOdtEntrySizeStim": dri["maxOdtEntrySizeStim"],
            "granularityOdtEntrySizeDaq": dri["granularityOdtEntrySizeDaq"],
            "granularityOdtEntrySizeStim": dri["granularityOdtEntrySizeStim"],
            "timestampMode": {
                "unit": dri["timestampMode"]["unit"],
                "fixed": dri["timestampMode"]["fixed"],
                "size": dri["timestampMode"]["size"],
            },
        }
        result["resolution"] = resolutionInfo
        channels = []
        daq_events = []
        if include_event_lists:
            for ecn in range(dpi.maxEventChannel):
                eci = self.getDaqEventInfo(ecn)
                cycle = eci["eventChannelTimeCycle"]
                maxDaqList = eci["maxDaqList"]
                priority = eci["eventChannelPriority"]
                time_unit = eci["eventChannelTimeUnit"]
                consistency = eci["daqEventProperties"]["consistency"]
                daq_supported = eci["daqEventProperties"]["daq"]
                stim_supported = eci["daqEventProperties"]["stim"]
                packed_supported = eci["daqEventProperties"]["packed"]
                name = self.fetch(eci.eventChannelNameLength)
                if name:
                    name = decode_bytes(name)
                channel = {
                    "name": name,
                    "priority": eci["eventChannelPriority"],
                    "unit": eci["eventChannelTimeUnit"],
                    "cycle": eci["eventChannelTimeCycle"],
                    "maxDaqList": eci["maxDaqList"],
                    "properties": {
                        "consistency": consistency,
                        "daq": daq_supported,
                        "stim": stim_supported,
                        "packed": packed_supported,
                    },
                }
                daq_event_info = DaqEventInfo(
                    name,
                    types.EVENT_CHANNEL_TIME_UNIT_TO_EXP[time_unit],
                    cycle,
                    maxDaqList,
                    priority,
                    consistency,
                    daq_supported,
                    stim_supported,
                    packed_supported,
                )
                daq_events.append(daq_event_info)
                channels.append(channel)
        result["channels"] = channels
        self.stim.setDaqEventInfo(daq_events)
        return result

    def getCurrentProtectionStatus(self):
        """"""
        if self.currentProtectionStatus is None:
            status = self.getStatus()
            self._setProtectionStatus(status.resourceProtectionStatus)
        return self.currentProtectionStatus

    def _setProtectionStatus(self, protection):
        """"""
        self.currentProtectionStatus = {
            "dbg": protection.dbg,
            "pgm": protection.pgm,
            "stim": protection.stim,
            "daq": protection.daq,
            "calpag": protection.calpag,
        }

    def cond_unlock(self, resources=None):
        """Conditionally unlock resources, i.e. only unlock locked resources.

        Precondition: Parameter "SEED_N_KEY_DLL" must be present and point to a valid DLL/SO.

        Parameters
        ----------
        resources: str
            Comma or space separated list of resources, e.g. "DAQ, CALPAG".
            The names are not case-sensitive.
            Valid identifiers are: "calpag", "daq", "dbg", "pgm", "stim".

            If omitted, try to unlock every available resource.

        Raises
        ------
        ValueError
            Invalid resource name.

        `dllif.SeedNKeyError`
            In case of DLL related issues.
        """
        import re

        from pyxcp.dllif import SeedNKeyError, SeedNKeyResult, getKey

        MAX_PAYLOAD = self.slaveProperties["maxCto"] - 2

        protection_status = self.getCurrentProtectionStatus()
        if any(protection_status.values()) and (not (self.seed_n_key_dll or self.seed_n_key_function)):
            raise RuntimeError("Neither seed-and-key DLL nor function specified, cannot proceed.")  # TODO: ConfigurationError
        if resources is None:
            result = []
            if self.slaveProperties["supportsCalpag"]:
                result.append("calpag")
            if self.slaveProperties["supportsDaq"]:
                result.append("daq")
            if self.slaveProperties["supportsStim"]:
                result.append("stim")
            if self.slaveProperties["supportsPgm"]:
                result.append("pgm")
            resources = ",".join(result)
        resource_names = [r.lower() for r in re.split(r"[ ,]", resources) if r]
        for name in resource_names:
            if name not in types.RESOURCE_VALUES:
                raise ValueError(f"Invalid resource name {name!r}.")
            if not protection_status[name]:
                continue
            resource_value = types.RESOURCE_VALUES[name]
            result = self.getSeed(types.XcpGetSeedMode.FIRST_PART, resource_value)
            seed = list(result.seed)
            length = result.length
            if length == 0:
                continue
            if length > MAX_PAYLOAD:
                remaining = length - len(seed)
                while remaining > 0:
                    result = self.getSeed(types.XcpGetSeedMode.REMAINING, resource_value)
                    seed.extend(list(result.seed))
                    remaining -= result.length
            self.logger.debug(f"Got seed {seed!r} for resource {resource_value!r}.")
            if self.seed_n_key_function:
                key = self.seed_n_key_function(resource_value, bytes(seed))
                self.logger.debug(f"Using seed and key function {self.seed_n_key_function.__name__!r}().")
                result = SeedNKeyResult.ACK
            elif self.seed_n_key_dll:
                self.logger.debug(f"Using seed and key DLL {self.seed_n_key_dll!r}.")
                result, key = getKey(
                    self.logger,
                    self.config.custom_dll_loader,
                    self.seed_n_key_dll,
                    resource_value,
                    bytes(seed),
                    self.seed_n_key_dll_same_bit_width,
                )
            if result == SeedNKeyResult.ACK:
                key = list(key)
                self.logger.debug(f"Unlocking resource {resource_value!r} with key {key!r}.")
                remaining = len(key)
                while key:
                    data = key[:MAX_PAYLOAD]
                    key_len = len(data)
                    self.unlock(remaining, data)
                    key = key[MAX_PAYLOAD:]
                    remaining -= key_len
            else:
                raise SeedNKeyError(f"SeedAndKey DLL returned: {SeedNKeyResult(result).name!r}")

    def identifier(self, id_value: int) -> str:
        """Return the identifier for the given value.
        Use this method instead of calling `getId()` directly.

        Parameters
        ----------
        id_value: int
            For standard identifiers, use the constants from `pyxcp.types.XcpGetIdType`.

        Returns
        -------
        str
        """
        gid = self.getId(id_value)
        if (gid.mode & 0x01) == 0x01:
            value = bytes(gid.identification or b"")
        else:
            value = self.fetch(gid.length)
        return decode_bytes(value)

    def id_scanner(self, scan_ranges: Optional[Collection[Collection[int]]] = None) -> Dict[str, str]:
        """Scan for available standard identification types (GET_ID).

        Parameters
        ----------
        scan_ranges: Optional[Collection[Collection[int]]]

        - If parameter is omitted or `None` test every standard identification type (s. GET_ID service)
          plus extensions by Vector Informatik.
        - Else `scan_ranges` must be a list-of-list.
            e.g: [[12, 80], [123], [240, 16, 35]]
                - The first list is a range (closed interval).
                - The second is a single value.
                - The third is a value list.

        Returns
        -------
        Dict[str, str]

        """
        result = {}

        def make_generator(sr):
            STD_IDS = {int(v): k for k, v in types.XcpGetIdType.__members__.items()}
            if sr is None:
                scan_range = STD_IDS.keys()
            else:
                scan_range = []
                if not isinstance(sr, Collection):
                    raise TypeError("scan_ranges must be of type `Collection`")
                for element in sr:
                    if not isinstance(element, Collection):
                        raise TypeError("scan_ranges elements must be of type `Collection`")
                    if not element:
                        raise ValueError("scan_ranges elements cannot be empty")
                    if len(element) == 1:
                        scan_range.append(element[0])  # Single value
                    elif len(element) == 2:
                        start, stop = element  # Value range
                        scan_range.extend(list(range(start, stop + 1)))
                    else:
                        scan_range.extend(element)  # Value list.
            scan_range = sorted(frozenset(scan_range))

            def generate():
                for idx, id_value in enumerate(scan_range):
                    if id_value in STD_IDS:
                        name = STD_IDS[id_value]
                    else:
                        name = f"USER_{idx}"
                    yield id_value, name,

            return generate()

        gen = make_generator(scan_ranges)
        for id_value, name in gen:
            status, response = self.try_command(self.identifier, id_value)
            if status == types.TryCommandResult.OK and response:
                result[name] = response
            elif status == types.TryCommandResult.XCP_ERROR and response.error_code == types.XcpError.ERR_CMD_UNKNOWN:
                break  # Nothing to do here.
            elif status == types.TryCommandResult.OTHER_ERROR:
                raise RuntimeError(f"Error while scanning for ID {id_value}: {response!r}")
        return result

    @property
    def start_datetime(self) -> int:
        """"""
        return self.transport.start_datetime

    def try_command(self, cmd: Callable, *args, **kws) -> Tuple[types.TryCommandResult, Any]:
        """Call master functions and handle XCP errors more gracefuly.

        Parameter
        ---------
        cmd: Callable
        args: list
            variable length arguments to `cmd`.
        kws: dict
            keyword arguments to `cmd`.

            `extra_msg`: str
                Additional info to log message (not passed to `cmd`).

        Returns
        -------

        Note
        ----
        Mainly used for plug-and-play applications, e.g. `id_scanner` may confronted with `ERR_OUT_OF_RANGE` errors, which
        is normal for this kind of applications -- or to test for optional commands.
        Use carefuly not to hide serious error causes.
        """
        try:
            extra_msg: Optional[str] = kws.get("extra_msg")
            if extra_msg:
                kws.pop("extra_msg")
            else:
                extra_msg = ""
            silent: Optional[bool] = kws.get("silent")
            if silent:
                kws.pop("silent")
            else:
                silent = False
            res = cmd(*args, **kws)
        except SystemExit as e:
            # print(f"\tUnexpected error while executing command {cmd.__name__!r}: {e!r}")
            if e.error_code == types.XcpError.ERR_CMD_UNKNOWN:
                # This is a rather common use-case, so let the user know that there is some functionality missing.
                if not silent:
                    if extra_msg:
                        self.logger.warning(f"Optional command {cmd.__name__!r} not implemented -- {extra_msg!r}")
                    else:
                        self.logger.warning(f"Optional command {cmd.__name__!r} not implemented.")
                return (types.TryCommandResult.NOT_IMPLEMENTED, e)
            else:
                return (types.TryCommandResult.XCP_ERROR, e)
        except Exception as e:
            return (types.TryCommandResult.OTHER_ERROR, e)
        else:
            return (types.TryCommandResult.OK, res)


def ticks_to_seconds(ticks, resolution):
    """Convert DAQ timestamp/tick value to seconds.

    Parameters
    ----------
    ticks: int

    unit: `GetDaqResolutionInfoResponse` as returned by :meth:`getDaqResolutionInfo`
    """
    warnings.warn(
        "ticks_to_seconds() deprecated, use factory :func:`make_tick_converter` instead.",
        Warning,
        stacklevel=1,
    )
    return (10 ** types.DAQ_TIMESTAMP_UNIT_TO_EXP[resolution.timestampMode.unit]) * resolution.timestampTicks * ticks


def make_tick_converter(resolution):
    """Make a function that converts tick count from XCP slave to seconds.

    Parameters
    ----------
    resolution: `GetDaqResolutionInfoResponse` as returned by :meth:`getDaqResolutionInfo`

    """
    exponent = types.DAQ_TIMESTAMP_UNIT_TO_EXP[resolution.timestampMode.unit]
    tick_resolution = resolution.timestampTicks
    base = (10**exponent) * tick_resolution

    def ticks_to_seconds(ticks):
        """Convert DAQ timestamp/tick value to seconds.

        Parameters
        ----------
        ticks: int

        Returns
        -------
        float
        """
        return base * ticks

    return ticks_to_seconds



#!/usr/bin/env python
"""Scan for available IDs.
"""

from pyxcp.cmdline import ArgumentParser


def main():
    ap = ArgumentParser(description="Scan for available IDs.")
    with ap.run() as x:
        x.connect()
        result = x.id_scanner()
        for key, value in result.items():
            print(f"{key}: {value}", end="\n\n")
        x.disconnect()


if __name__ == "__main__":
    main()
